# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import os
from typing import Any, Dict, Literal, Optional
from mito_ai.utils.version_utils import MITOSHEET_HELPER_PRIVATE, is_pro
from mito_ai.utils.schema import UJ_AI_MITO_API_NUM_USAGES, UJ_MITOSHEET_TELEMETRY, UJ_STATIC_USER_ID, UJ_USER_EMAIL, UJ_FEEDBACKS_V2
from mito_ai.utils.db import get_user_field
from mito_ai._version import __version__
from mito_ai.utils.utils import is_running_test
from mito_ai.completions.models import MessageType
import analytics

WRITE_KEY = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 
analytics.write_key = WRITE_KEY

# If you want, you can optionally choose to print logs
# helpful for debugging.
PRINT_LOGS = False

################################# 
# Mito AI Completion 
# Constants for logging the success or error of Mito AI
MITO_AI_COMPLETION_SUCCESS = 'mito_ai_success'
MITO_AI_COMPLETION_ERROR = 'mito_ai_error'

# Params 
# - logging the type of key 
KEY_TYPE_PARAM = 'AI_key_type'
MITO_SERVER_KEY: Literal['mito_server_key'] = 'mito_server_key'
USER_KEY: Literal['user_key'] = 'user_key'
# - logging if the user is in dev mode
IS_DEV_MODE_PARAM = 'is_dev_mode'

# - logging the number of usages of the Mito server
MITO_SERVER_NUM_USAGES = 'mito_server_num_usages'
#################################

################################# 
# Mito Server Free Tier Reached
MITO_SERVER_FREE_TIER_LIMIT_REACHED = 'mito_server_free_tier_limit_reached'
#################################

def is_dev_mode() -> bool:
    """
    Check if mito-ai is installed in editable/development mode.
    
    This function detects editable installs using the modern PEP 660 standard
    (pip >= 21.3). Works for most development scenarios where developers use
    `pip install -e .`
    
    Returns:
        bool: True if running in development mode, False otherwise.
        
    Limitations:
        - Requires pip >= 21.3 for reliable detection
        - Won't detect manual PYTHONPATH manipulation
        - Won't detect legacy .egg-link installations (very old pip)
        
    Note: This is a best-effort detection. For 100% reliability, consider
    also setting a MITO_DEVELOPER_MODE environment variable.
    """
    try:
        import importlib.metadata
        import json
        
        dist = importlib.metadata.distribution('mito-ai')
        direct_url_text = dist.read_text('direct_url.json')
        if direct_url_text:
            direct_url = json.loads(direct_url_text)
            direct_url_dir_info = direct_url.get('dir_info', {})
            editable = direct_url_dir_info.get('editable', None)
            return False if editable is None else editable
    except Exception:
        pass
    
    return False

def telemetry_turned_on(key_type: Optional[str] = None) -> bool:
    """
    Helper function that tells you if logging is turned on or
    turned off on the entire Mito instance
    """
    
    # If the user is on the Mito server, then they are sending
    # us their information already
    if key_type == 'mito_server_key':
        return True
    
    # If private helper is installed, then we don't log anything
    if MITOSHEET_HELPER_PRIVATE:
        return False

    # TODO: Check if the an enterprise user has turned telemetry to true

    # If Mito Pro is on, then don't log anything
    if is_pro():
        return False

    telemetry = get_user_field(UJ_MITOSHEET_TELEMETRY)
    if telemetry is None:
        return False
    
    return bool(telemetry)

def identify(key_type: Optional[str] = None) -> None:
    """
    Helper function for identifying a user. We just take
    their python version, mito version, and email.
    """
    if not telemetry_turned_on(key_type):
        return

    static_user_id = get_user_field(UJ_STATIC_USER_ID)
    user_email = get_user_field(UJ_USER_EMAIL)
    feedbacks_v2 = get_user_field(UJ_FEEDBACKS_V2)

    params = {
        'version_mitoai': __version__,
        'email': user_email,
        'is_pro': is_pro(),
        'is_jupyterhub': 'True' if 'JUPYTERHUB_API_URL' in os.environ else 'False',
        'is_mito_jupyterhub': 'True' if os.getenv('MITO_JUPYTERHUB') is not None else 'False',
        IS_DEV_MODE_PARAM: is_dev_mode(),
        UJ_FEEDBACKS_V2: feedbacks_v2
    }

    if not is_running_test():
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        try:
            analytics.identify(static_user_id, params)
        except Exception as e:
            pass

def chunk_param(param: str, param_name: str, chunk_size: int=250) -> Dict[str, str]:
    """
    Split a string into chunks of 250 characters.
    
    Args:
        param: The string to be chunked
        param_name: The name of the param to be chunked (used as prefix for the chunked keys)
        chunk_size: The number of characters in each chunk

    Returns:
        dict: A dictionary with keys 'response_part_1', 'response_part_2', etc.
    """

    chunks = {}

    if not param:
        return {}
    
    num_chunks = (len(param) + chunk_size - 1) // chunk_size

    for i in range(num_chunks):
        start = i * chunk_size
        end = min(start + chunk_size, len(param))
        chunks[f'{param_name}_part_{i + 1}'] = param[start:end]

    return chunks

def log(
        log_event: str, 
        params: Optional[Dict[str, Any]]=None, 
        error: Optional[BaseException]=None, 
        key_type: Optional[Literal['mito_server_key', 'user_key']] = None,
        thread_id: Optional[str] = None
    ) -> None:
    """
    This function is the entry point for all logging. 

    If telemetry is not turned off and we are not running tests,
    we log the ai event
    """

    final_params: Dict[str, Any] = params or {}
    
    # Then, make sure to add the user email
    final_params['email'] = get_user_field(UJ_USER_EMAIL)

    # Add the error if it exists
    if error is not None:
        final_params['error'] = str(error)

    if thread_id is not None:
        final_params['thread_id'] = thread_id

    # Finally, do the acutal logging. We do not log anything when tests are
    # running, or if telemetry is turned off
    if not is_running_test() and telemetry_turned_on(key_type):
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        
        try:
            analytics.track(
                get_user_field(UJ_STATIC_USER_ID), 
                log_event, 
                final_params
            )
        except Exception as e:
            pass
        

    # If we want to print the logs for debugging reasons, then we print them as well
    if PRINT_LOGS:
        print(
            log_event, 
            final_params
        )

    # TODO: Eventually we want to hook this up to the mito log uploader 
    # so enterprises can log usage if they want to.

def log_ai_completion_success(
    key_type: Literal['mito_server_key', 'user_key'],
    message_type: MessageType,
    last_message_content: str,
    user_input: str,
    response: Dict[str, Any],
    thread_id: str,
    model: str
) -> None:
    """
    Logs AI completion success based on the input location.

    Args:
        key_type: The type of key that was used to get the AI completion
        message_type: The type of message that was sent to the AI
        last_message_content: The last message sent to the AI
        user_input: The user input that was sent to the AI
        response: The response received from the AI
        thread_id: The thread ID for the conversation
        model: The model that was used to get the AI completion
    """

    # Params that every log has
    base_params = {
        KEY_TYPE_PARAM: str(key_type),
        IS_DEV_MODE_PARAM: is_dev_mode(),
        'model': model,
    }

    try:
        code_cell_input = json.dumps(
            last_message_content.split("Code in the active code cell:")[-1]
            .strip()
            .split("```python")[1]
            .strip()
            .split("```")[0]
        )
        
        num_usages = get_user_field(UJ_AI_MITO_API_NUM_USAGES)
    except:
        # Most user prompts will have an associated code cell that serves as the input context.
        # However, types like agent:planning (RIP) do not have a code cell input.
        code_cell_input = ""
        num_usages = -1

    # Chunk certain params to work around mixpanel's 255 character limit
    code_cell_input_chunks = chunk_param(code_cell_input, "code_cell_input")
    response_chunks = chunk_param(response["completion"], "response")

    for chunk_key, chunk_value in code_cell_input_chunks.items():
        base_params[chunk_key] = chunk_value

    for chunk_key, chunk_value in response_chunks.items():
        base_params[chunk_key] = chunk_value

    # Log number of usages (for mito server)
    if num_usages is not None:
        base_params[MITO_SERVER_NUM_USAGES] = str(num_usages)

    if message_type == MessageType.SMART_DEBUG:
        error_message = (
            last_message_content.split("Error Message:")[-1]
            .split("ERROR ANALYSIS:")[0]
            .strip()
        )
        error_type = error_message.split(": ")[0]

        final_params = base_params
        final_params["error_message"] = error_message
        final_params["error_type"] = error_type

        log("mito_ai_smart_debug_success", params=final_params, key_type=key_type, thread_id=thread_id)
    elif message_type == MessageType.CODE_EXPLAIN:
        final_params = base_params

        log("mito_ai_code_explain_success", params=final_params, key_type=key_type, thread_id=thread_id)
    elif message_type == MessageType.CHAT:
        final_params = base_params

        # Chunk the user input
        user_input_chunks = chunk_param(user_input, "user_input")
        
        for chunk_key, chunk_value in user_input_chunks.items():
            final_params[chunk_key] = chunk_value

        log("mito_ai_chat_success", params=final_params, key_type=key_type, thread_id=thread_id)
    elif message_type == MessageType.AGENT_EXECUTION:
        final_params = base_params

        # Chunk the user input
        user_input_chunks = chunk_param(user_input, "user_input")
        
        for chunk_key, chunk_value in user_input_chunks.items():
            final_params[chunk_key] = chunk_value
            
        # If the user input is not empty, then this is the user giving the agent a new task
        new_user_input = 'True' if len(user_input_chunks) > 0 else 'False'
        final_params["new_user_input"] = new_user_input

        log("mito_ai_agent_execution_success", params=final_params, key_type=key_type, thread_id=thread_id)
    elif message_type == MessageType.INLINE_COMPLETION:
        final_params = base_params
        log("mito_ai_inline_completion_success", params=final_params, key_type=key_type, thread_id=thread_id)
    elif message_type == MessageType.AGENT_AUTO_ERROR_FIXUP:
        final_params = base_params
        log("mito_ai_agent_auto_error_fixup_success", params=final_params, key_type=key_type, thread_id=thread_id)
    else:
        final_params = base_params
        final_params["note"] = (
            "This input_location has not been accounted for in `telemetry_utils.py`."
        )

        log(f"mito_ai_{message_type.value}_success", params=final_params, key_type=key_type, thread_id=thread_id)
        
def log_db_connection_attempt(connection_type: str) -> None:
    log("mito_ai_db_connection_attempt", params={"connection_type": connection_type})

def log_db_connection_success(connection_type: str, schema: Dict[str, Any]) -> None:
    log(
        "mito_ai_db_connection_success",
        params={
            "connection_type": connection_type,
        },
    )

def log_db_connection_error(connection_type: str, error_message: str) -> None:
    log(
        "mito_ai_db_connection_error", 
        params={
            "connection_type": connection_type, 
            "error_message": error_message,
        }
    )



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from dataclasses import dataclass
from enum import Enum
from typing import Literal, Optional


class MessageType(str, Enum):
    """Types of app builder messages."""
    BUILD_APP = "build-app"


@dataclass(frozen=True)
class AppBuilderError:
    """Error information for app builder operations."""
    
    # Error type.
    error_type: str
    
    # Error title.
    title: str
    
    # Error traceback.
    traceback: Optional[str] = None
    
    # Hint to fix the error.
    hint: Optional[str] = None
    
    @classmethod
    def from_exception(cls, e: Exception, hint: Optional[str] = None) -> "AppBuilderError":
        """Create an error from an exception.

        Args:
            e: The exception.
            hint: Optional hint to fix the error.

        Returns:
            The app builder error.
        """
        return cls(
            error_type=type(e).__name__,
            title=str(e),
            traceback=getattr(e, "__traceback__", None) and str(e.__traceback__),
            hint=hint,
        )


@dataclass(frozen=True)
class ErrorMessage(AppBuilderError):
    """Error message."""
    
    # Message type.
    type: Literal["error"] = "error"


@dataclass(frozen=True)
class BuildAppRequest:
    """Request to build an app."""
    
    # Request type.
    type: Literal["build-app"]
    
    # Message ID.
    message_id: str
    
    # Path to the app file.
    path: str


@dataclass(frozen=True)
class BuildAppReply:
    """Reply to a build app request."""
    
    # ID of the request message this is replying to.
    parent_id: str
    
    # URL of the deployed app.
    url: str
    
    # Optional error information.
    error: Optional[AppBuilderError] = None
    
    # Type of reply.
    type: Literal["build-app"] = "build-app" 


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
import time
import logging
from typing import Any, Union
import zipfile
import tempfile
from mito_ai.utils.create import initialize_user
from mito_ai.utils.websocket_base import BaseWebSocketHandler
from mito_ai.app_builder.models import (
    BuildAppReply,
    AppBuilderError,
    ErrorMessage,
    MessageType
)
from mito_ai.logger import get_logger
import requests

# API endpoint for getting pre-signed URL
API_BASE_URL = "https://fr12uvtfy5.execute-api.us-east-1.amazonaws.com"

class AppBuilderHandler(BaseWebSocketHandler):
    """Handler for app building requests."""
    
    def initialize(self) -> None:
        """Initialize the WebSocket handler."""
        super().initialize()
        self.log.debug("Initializing app builder websocket connection %s", self.request.path)
    
    @property
    def log(self) -> logging.Logger:
        """Use Mito AI logger."""
        return get_logger()
    
    async def get(self, *args: Any, **kwargs: Any) -> None:
        """Get an event to open a socket or check service availability."""
        # Check if this is just a service availability check
        if self.get_query_argument('check_availability', None) == 'true':
            self.set_status(200)
            self.finish()
            return

        await super().pre_get()  # Authenticate and authorize
        initialize_user()  # Initialize user directory structure
        
        reply = super().get(*args, **kwargs)
        if reply is not None:
            await reply
    
    async def on_message(self, message: Union[str, bytes]) -> None:
        """Handle incoming messages on the WebSocket.
        
        Args:
            message: The message received on the WebSocket.
        """
        start = time.time()
        
        # Convert bytes to string if needed
        if isinstance(message, bytes):
            message = message.decode('utf-8')
        
        self.log.debug("App builder message received: %s", message)
        
        try:
            parsed_message = self.parse_message(message)
            message_type = parsed_message.get('type')
            
            if message_type == MessageType.BUILD_APP.value:
                # Handle build app request
                await self._handle_build_app(parsed_message)
            else:
                self.log.error(f"Unknown message type: {message_type}")
                error = AppBuilderError(
                    error_type="InvalidRequest",
                    title=f"Unknown message type: {message_type}"
                )
                self.reply(ErrorMessage(**error.__dict__))
                
        except ValueError as e:
            self.log.error("Invalid app builder request", exc_info=e)
            error = AppBuilderError.from_exception(e)
            self.reply(ErrorMessage(**error.__dict__))
        except Exception as e:
            self.log.error("Error handling app builder message", exc_info=e)
            error = AppBuilderError.from_exception(
                e, 
                hint="An error occurred while building the app. Please check the logs for details."
            )
            self.reply(ErrorMessage(**error.__dict__))
            
        latency_ms = round((time.time() - start) * 1000)
        self.log.info(f"App builder handler processed in {latency_ms} ms.")
    
    async def _handle_build_app(self, message: dict) -> None:
        """Handle a build app request.
        
        Args:
            message: The parsed message.
        """
        message_id = message.get('message_id', '')  # Default to empty string if not present
        app_path = message.get('path')
        
        if not message_id:
            self.log.error("Missing message_id in request")
            return
        
        if not app_path:
            error = AppBuilderError(
                error_type="InvalidRequest",
                title="Missing 'path' parameter"
            )
            self.reply(BuildAppReply(
                parent_id=message_id,
                url="",
                error=error
            ))
            return
        
        try:
            # This is a placeholder for the actual app building logic
            # In a real implementation, this would deploy the app to a hosting service
            # and return the URL
            deploy_url = await self._deploy_app(app_path)
            
            # Send the response
            self.reply(BuildAppReply(
                parent_id=message_id,
                url=deploy_url
            ))
            
        except Exception as e:
            self.log.error(f"Error building app: {e}", exc_info=e)
            error = AppBuilderError.from_exception(e)
            self.reply(BuildAppReply(
                parent_id=message_id,
                url="",
                error=error
            ))
    
    async def _deploy_app(self, app_path: str) -> str:
        """Deploy the app using pre-signed URLs.
        
        Args:
            app_path: Path to the app file.
            
        Returns:
            The URL of the deployed app.
        """
        # Get app name from the path
        app_name = os.path.basename(app_path).split('.')[0]
        self.log.info(f"Deploying app: {app_name} from path: {app_path}")
        
        try:
            # Step 1: Get pre-signed URL from API
            self.log.info("Getting pre-signed upload URL...")
            url_response = requests.get(f"{API_BASE_URL}/get-upload-url?app_name={app_name}")
            url_response.raise_for_status()
            
            url_data = url_response.json()
            presigned_url = url_data['upload_url']
            expected_app_url = url_data['expected_app_url']
            
            self.log.info(f"Received pre-signed URL. App will be available at: {expected_app_url}")
            
            # Step 2: Create a zip file of the app.
            temp_zip_path = None
            try:
                # Create temp file and close it before writing to avoid file handle conflicts
                with tempfile.NamedTemporaryFile(suffix='.zip', delete=False) as temp_zip:
                    temp_zip_path = temp_zip.name

                self.log.info("Zipping application files...")
                with zipfile.ZipFile(temp_zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
                    for root, _, files in os.walk(app_path):
                        for file in files:
                            file_path = os.path.join(root, file)
                            zipf.write(file_path, arcname=os.path.relpath(file_path, app_path))

                upload_response = await self._upload_app_to_s3(temp_zip_path, presigned_url)
            except Exception as e:
                self.log.error(f"Error zipping app: {e}")
                raise
            finally:
                # Clean up
                if temp_zip_path is not None:
                    os.remove(temp_zip_path)
            
            self.log.info(f"Upload successful! Status code: {upload_response.status_code}")
            
            self.log.info(f"Deployment initiated. App will be available at: {expected_app_url}")
            return expected_app_url # type: ignore
            
        except requests.exceptions.RequestException as e:
            self.log.error(f"Error during API request: {e}")
            if hasattr(e, 'response') and e.response is not None:
                try:
                    error_detail = e.response.json()
                    self.log.error(f"Server error details: {error_detail}")
                except:
                    self.log.error(f"Server response: {e.response.text}")
            raise Exception(f"Deployment failed: {str(e)}")
        except Exception as e:
            self.log.error(f"Error during deployment: {str(e)}")
            raise

    async def _upload_app_to_s3(self, app_path: str, presigned_url: str) -> requests.Response:
        """Upload the app to S3 using the presigned URL."""
        with open(app_path, 'rb') as file_data:
            upload_response = requests.put(
                presigned_url,
                data=file_data,
                headers={'Content-Type': 'application/zip'}
            )
            upload_response.raise_for_status()
            
        return upload_response


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import os
import uuid
import time
import random
from datetime import datetime, timedelta

def generate_mito_chat_file(number, base_time):
    """Generate a single Mito chat JSON file with the specified print number."""
    
    # Create unique timestamps with some variation
    creation_ts = base_time - random.uniform(0, 3600 * 24 * 7)  # Random time within the last week
    last_interaction_ts = creation_ts + random.uniform(300, 1800)  # 5-30 minutes after creation
    
    # Generate a unique thread ID
    thread_id = str(uuid.uuid4())
    
    # Create the file content
    file_content = {
        "thread_id": thread_id,
        "creation_ts": creation_ts,
        "last_interaction_ts": last_interaction_ts,
        "name": f"Print {number}",
        "ai_optimized_history": [
            {
                "role": "system",
                "content": "You are Mito Data Copilot, an AI assistant for Jupyter. You're a great python programmer, a seasoned data scientist and a subject matter expert."
            },
            {
                "role": "user",
                "content": f"""You are an expert python programmer writing a script in a Jupyter notebook. You are given a set of variables, existing code, and a task.

There are two possible types of responses you might give:
1. Code Update: If the task requires modifying or extending the existing code, respond with the updated active code cell and a short explanation of the changes made. 
2. Explanation/Information: If the task does not require a code update, provide an explanation, additional information about a package, method, or general programming question, without writing any code. Keep your response concise and to the point.

When responding:
- Do not use the word "I"
- Do not recreate variables that already exist
- Keep as much of the original code as possible

<Example>

Files in the current directory:
file_name: sales.csv

Defined Variables:
{{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

Code in the active code cell:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
```

Your task: convert the transaction_date column to datetime and then multiply the total_price column by the sales_multiplier.

Output:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
sales_df['transaction_date'] = pd.to_datetime(sales_df['transaction_date'])
sales_df['total_price'] = sales_df['total_price'] * sales_multiplier
```

Converted the `transaction_date` column to datetime using the built-in pd.to_datetime function and multiplied the `total_price` column by the `sales_multiplier` variable.

</Example>

Files in the current directory:
{{'file_name': 'used_cars.csv'}}

Defined Variables:
{{'variable_name': 'mitosheet', 'type': "<class 'module'>", 'value': "<module 'mitosheet' from '/Users/aarondiamond-reivich/Mito/mito/mito-ai/venv/lib/python3.11/site-packages/mitosheet/__init__.py'>"}}
{{'variable_name': 'pd', 'type': "<class 'module'>", 'value': "<module 'pandas' from '/Users/aarondiamond-reivich/Mito/mito/mito-ai/venv/lib/python3.11/site-packages/pandas/__init__.py'>"}}
{{'variable_name': 'uuid', 'type': "<class 'module'>", 'value': "<module 'uuid' from '/Library/Frameworks/Python.framework/Versions/3.11/lib/python3.11/uuid.py'>"}}

Code in the active code cell:
```python

```

Your task: print {number}
"""
            },
            {
                "role": "assistant",
                "content": f"""```python
print({number})
```

The task was to print the integer {number}. The `print` function is used to output the value."""
            }
        ],
        "display_history": [
            {
                "role": "system",
                "content": "You are Mito Data Copilot, an AI assistant for Jupyter. You're a great python programmer, a seasoned data scientist and a subject matter expert."
            },
            {
                "role": "user",
                "content": f"```python```print {number}"
            },
            {
                "role": "assistant",
                "content": f"""```python
print({number})
```

The task was to print the integer {number}. The `print` function is used to output the value."""
            }
        ],
        "chat_history_version": 2
    }
    
    return file_content

def main(num_files=30):
    """Generate the specified number of Mito chat files."""
    
    # Create the directory if it doesn't exist
    directory = os.path.expanduser("~/.mito/ai-chats")
    os.makedirs(directory, exist_ok=True)
    
    current_time = time.time()
    
    for i in range(1, num_files + 1):
        file_content = generate_mito_chat_file(i, current_time)
        
        # Generate a unique filename
        filename = f"print_{i}_{str(uuid.uuid4())[:8]}.json"
        file_path = os.path.join(directory, filename)
        
        # Write the file
        with open(file_path, 'w') as f:
            json.dump(file_content, f, indent=2)
        
        print(f"Created file: {file_path}")
    
    print(f"\nSuccessfully created {num_files} Mito chat files in: {directory}")

if __name__ == "__main__":
    # You can change the number of files to generate
    main(30)  # Default: Generate 30 files


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from dataclasses import dataclass
import json
from typing import Any, Final, Union
import tornado
import os
from jupyter_server.base.handlers import APIHandler
from mito_ai.rules.utils import RULES_DIR_PATH, get_all_rules, get_rule, set_rules_file


class RulesHandler(APIHandler):
    """Handler for operations on a specific setting"""
    
    @tornado.web.authenticated
    def get(self, key: Union[str, None] = None) -> None:
        """Get a specific rule by key or all rules if no key provided"""
        if key is None or key == '':
            # No key provided, return all rules
            rules = get_all_rules()
            self.finish(json.dumps(rules))
        else:
            # Key provided, return specific rule
            rule_content = get_rule(key)
            if rule_content is None:
                self.set_status(404)
                self.finish(json.dumps({"error": f"Rule with key '{key}' not found"}))
            else:
                self.finish(json.dumps({"key": key, "content": rule_content}))
    
    @tornado.web.authenticated
    def put(self, key: str) -> None:
        """Update or create a specific setting"""
        data = json.loads(self.request.body)
        if 'content' not in data:
            self.set_status(400)
            self.finish(json.dumps({"error": "Content is required"}))
            return
            
        set_rules_file(key, data['content'])
        self.finish(json.dumps({"status": "updated", "rules file ": key}))





# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import os
from mito_ai.db.crawlers import snowflake, base_crawler
from mito_ai.db.crawlers.constants import SUPPORTED_DATABASES


def setup_database_dir(
    db_dir_path: str, connections_path: str, schemas_path: str
) -> None:
    """
    Setup the database directory.
    """

    # Ensure the db directory exists
    os.makedirs(db_dir_path, exist_ok=True)

    # Create connections.json if it doesn't exist
    if not os.path.exists(connections_path):
        with open(connections_path, "w") as f:
            json.dump({}, f, indent=4)

    # Create schemas.json if it doesn't exist
    if not os.path.exists(schemas_path):
        with open(schemas_path, "w") as f:
            json.dump({}, f, indent=4)


def save_connection(
    connections_path: str, connection_id: str, connection_details: dict
) -> None:
    """
    Save a connection to the connections.json file.

    Args:
        connections_path (str): The path to the connections.json file.
        connection_id (str): The UUID of the connection to save.
        connection_details (dict): The details of the connection to save.
    """

    with open(connections_path, "r") as f:
        connections = json.load(f)

    # Add the new connection
    connections[connection_id] = connection_details

    # Write back to file
    with open(connections_path, "w") as f:
        json.dump(connections, f, indent=4)


def delete_connection(connections_path: str, connection_id: str) -> None:
    """
    Delete a connection by UUID.
    """

    # Read existing connections
    with open(connections_path, "r") as f:
        connections = json.load(f)

    # Remove the connection
    del connections[connection_id]

    # Write back to file
    with open(connections_path, "w") as f:
        json.dump(connections, f, indent=4)


def delete_schema(schemas_path: str, schema_id: str) -> None:
    """
    Delete a schema by UUID.

    Args:
        schemas_path (str): The path to the schemas.json file.
        schema_id (str): The UUID of the schema to delete.
    """

    with open(schemas_path, "r") as f:
        schemas = json.load(f)

    del schemas[schema_id]

    with open(schemas_path, "w") as f:
        json.dump(schemas, f, indent=4)


def crawl_and_store_schema(
    schemas_path: str,
    connection_id: str,
    connection_details: dict,
) -> dict:
    """
    Crawl and store schema for a given connection.

    Args:
        schemas_path (str): The path to the schemas.json file.
        connection_id (str): The UUID of the connection to crawl.
        username (str): The username for the connection.
        password (str): The password for the connection.
        account (str): The account for the connection.
        warehouse (str): The warehouse for the connection.

    Returns:
        tuple[bool, str]: A tuple containing a boolean indicating success and an error message.
    """
    if connection_details["type"] == "snowflake":
        schema = snowflake.crawl_snowflake(
            connection_details["username"],
            connection_details["password"],
            connection_details["account"],
            connection_details["warehouse"],
        )
    elif connection_details["type"] == "postgres":
        conn_str = f"postgresql+psycopg2://{connection_details['username']}:{connection_details['password']}@{connection_details['host']}:{connection_details['port']}/{connection_details['database']}"
        schema = base_crawler.crawl_db(conn_str, "postgres")
    elif connection_details["type"] == "sqlite":
        conn_str = f"sqlite:///{connection_details['database']}"
        schema = base_crawler.crawl_db(conn_str, "sqlite")
    elif connection_details["type"] == "mysql":
        conn_str = f"mysql+pymysql://{connection_details['username']}:{connection_details['password']}@{connection_details['host']}:{connection_details['port']}/{connection_details['database']}"
        schema = base_crawler.crawl_db(conn_str, "mysql")
    elif connection_details["type"] == "mssql":
        odbc_driver_version = connection_details["odbc_driver_version"]
        conn_str = f"mssql+pyodbc://{connection_details['username']}:{connection_details['password']}@{connection_details['host']}:{connection_details['port']}/{connection_details['database']}?driver=ODBC+Driver+{odbc_driver_version}+for+SQL+Server"
        schema = base_crawler.crawl_db(conn_str, "mssql")
    elif connection_details["type"] == "oracle":
        conn_str = f"oracle+oracledb://{connection_details['username']}:{connection_details['password']}@{connection_details['host']}:{connection_details['port']}?service_name={connection_details['service_name']}"
        schema = base_crawler.crawl_db(conn_str, "oracle")

    if schema["error"]:
        return {
            "success": False,
            "error_message": schema["error"],
            "schema": {},
        }

    # If we successfully crawled the schema, write it to schemas.json
    with open(schemas_path, "r+") as f:
        # Load the existing schemas
        schemas = json.load(f)
        # Remove the error key from the schema and add the crawled schema
        schema.pop("error", None)
        schemas[connection_id] = schema["schema"]
        # Move to the beginning of the file and write the new schema
        f.seek(0)
        json.dump(schemas, f, indent=4)
        f.truncate()
    return {
        "success": True,
        "error_message": "",
        "schema": schema,
    }


def install_db_drivers(db_type: str) -> dict:
    """
    Install required database drivers for the given database type.

    Args:
        db_type (str): The type of database (e.g. 'snowflake', 'postgres')

    Returns:
        dict: A dictionary containing success status and error message if any
    """
    from mito_ai.utils.utils import get_installed_packages, install_packages

    installed_packages = get_installed_packages()
    required_packages = SUPPORTED_DATABASES[db_type].get("drivers", [])
    packages_to_install = []

    for package in required_packages:
        if package not in installed_packages:
            packages_to_install.append(package)

    if len(packages_to_install) > 0:
        install_result = install_packages(packages_to_install)
        if not install_result["success"]:
            return {
                "success": False,
                "error": f"Failed to install {db_type} drivers: {install_result['error']}",
            }

    return {"success": True, "error": None}



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import logging

from traitlets.config import Application


_LOGGER = None  # type: logging.Logger | None


def get_logger() -> logging.Logger:
    """Create a logger for the Mito AI module.
    
    The logger will be attached as a child of the Jupyter server logger.
    This allows for easier filtering/flagging of log messages in the server logs.
    
    Example:

    The following snippet shows a log message produced by JupyterLab followed by
    two log messages produced by the Mito AI module.

        [D 2024-12-16 15:49:07.333 LabApp] 204 PUT /lab/api/workspaces/default?1734360547329 (ea7486428da24ff3921aebd4422611d9@::1) 1.85ms
        [D 2024-12-16 15:49:08.293 ServerApp.mito_ai] Message received: {...}
        [D 2024-12-16 15:49:08.293 ServerApp.mito_ai] Requesting completion from Mito server.

    You can filter the server logs to only show messages produced by the Mito AI module:

        jupyter lab --debug 2>&1 | egrep mito_ai
    """
    global _LOGGER
    if _LOGGER is None:
        app = Application.instance()
        _LOGGER = logging.getLogger("{!s}.mito_ai".format(app.log.name))
        Application.clear_instance()

    return _LOGGER



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
from mito_ai.anthropic_client import get_anthropic_system_prompt_and_messages, extract_and_parse_anthropic_json_response, AnthropicClient, ANTHROPIC_FAST_MODEL
from mito_ai.utils.anthropic_utils import get_anthropic_completion_function_params
from anthropic.types import MessageParam, Message, ContentBlock, TextBlock, ToolUseBlock, Usage
from openai.types.chat import ChatCompletionMessageParam, ChatCompletionUserMessageParam, ChatCompletionAssistantMessageParam, ChatCompletionSystemMessageParam
from mito_ai.completions.models import ResponseFormatInfo, AgentResponse
from unittest.mock import MagicMock, patch
import anthropic
from typing import List, Dict, Any, cast, Union

# Dummy base64 image (1x1 PNG)
DUMMY_IMAGE_DATA_URL = (
    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mP8/wcAAgMBAp9l9AAAAABJRU5ErkJggg=="
)

def test_mixed_text_and_image():
    messages: List[ChatCompletionMessageParam] = [
        ChatCompletionSystemMessageParam(role="system", content="You are a helpful assistant."),
        ChatCompletionUserMessageParam(role="user", content=[
            {"type": "text", "text": "Here is an image:"},
            {"type": "image_url", "image_url": {"url": DUMMY_IMAGE_DATA_URL}}
        ])
    ]
    system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
    
    assert system_prompt == "You are a helpful assistant."
    assert len(anthropic_messages) == 1
    message = anthropic_messages[0]
    assert message["role"] == "user"
    content = message["content"]
    assert isinstance(content, list)
    assert len(content) == 2
    
    # Check text content
    text_block = cast(Dict[str, str], content[0])
    assert text_block["type"] == "text"
    assert text_block["text"] == "Here is an image:"
    
    # Check image content
    image_block = cast(Dict[str, Dict[str, str]], content[1])
    assert image_block["type"] == "image"
    assert image_block["source"]["type"] == "base64"
    assert image_block["source"]["media_type"] == "image/png"

def test_no_system_instructions_only_content():
    messages: List[ChatCompletionMessageParam] = [
        ChatCompletionUserMessageParam(role="user", content="Hello!"),
        ChatCompletionAssistantMessageParam(role="assistant", content="Hi, how can I help you?")
    ]
    system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
    
    assert isinstance(system_prompt, anthropic.NotGiven)
    assert len(anthropic_messages) == 2
    assert anthropic_messages[0]["role"] == "user"
    assert anthropic_messages[0]["content"] == "Hello!"
    assert anthropic_messages[1]["role"] == "assistant"
    assert anthropic_messages[1]["content"] == "Hi, how can I help you?"

def test_system_instructions_and_content():
    messages: List[ChatCompletionMessageParam] = [
        ChatCompletionSystemMessageParam(role="system", content="You are a helpful assistant."),
        ChatCompletionUserMessageParam(role="user", content="What is the weather today?")
    ]
    system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
    
    assert system_prompt == "You are a helpful assistant."
    assert len(anthropic_messages) == 1
    assert anthropic_messages[0]["role"] == "user"
    assert anthropic_messages[0]["content"] == "What is the weather today?"

def test_multiple_system_messages():
    messages: List[ChatCompletionMessageParam] = [
        ChatCompletionSystemMessageParam(role="system", content="First system message."),
        ChatCompletionSystemMessageParam(role="system", content="Second system message."),
        ChatCompletionUserMessageParam(role="user", content="Hello!")
    ]
    system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
    
    # Should take the last system message
    assert system_prompt == "Second system message."
    assert len(anthropic_messages) == 1
    assert anthropic_messages[0]["role"] == "user"
    assert anthropic_messages[0]["content"] == "Hello!"

def test_empty_message_content():
    messages: List[ChatCompletionMessageParam] = [
        cast(ChatCompletionMessageParam, {"role": "user"}),  # Missing content
        ChatCompletionAssistantMessageParam(role="assistant", content="Hi!")
    ]
    system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
    
    assert isinstance(system_prompt, anthropic.NotGiven)
    assert len(anthropic_messages) == 1  # Should skip the message with missing content
    assert anthropic_messages[0]["role"] == "assistant"
    assert anthropic_messages[0]["content"] == "Hi!"

def test_extract_json_from_tool_use():
    # Create a mock response with tool use
    tool_use_block = ToolUseBlock(
        type="tool_use",
        id="test_id",
        name="agent_response",
        input={"key": "value"}
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[tool_use_block],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    result = extract_and_parse_anthropic_json_response(response)
    assert result == {"key": "value"}

def test_extract_json_from_text():
    # Create a mock response with JSON in text
    text_block = TextBlock(
        type="text",
        text='Here is some JSON: {"key": "value"}'
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[text_block],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    result = extract_and_parse_anthropic_json_response(response)
    assert result == {"key": "value"}

def test_extract_json_from_text_with_multiple_blocks():
    # Create a mock response with multiple text blocks
    text_block1 = TextBlock(
        type="text",
        text='Here is the JSON: {"key": "value"}'  # Put JSON in first block since that's what the implementation checks
    )
    text_block2 = TextBlock(
        type="text",
        text="Some text after JSON"
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[text_block1, text_block2],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    result = extract_and_parse_anthropic_json_response(response)
    assert result == {"key": "value"}

def test_invalid_json_in_text():
    # Create a mock response with invalid JSON in text
    text_block = TextBlock(
        type="text",
        text='Here is invalid JSON: {"key": value}'
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[text_block],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    with pytest.raises(Exception) as exc_info:
        extract_and_parse_anthropic_json_response(response)
    assert "No valid AgentResponse format found" in str(exc_info.value)

def test_no_json_in_text():
    # Create a mock response with no JSON in text
    text_block = TextBlock(
        type="text",
        text="This is just plain text with no JSON"
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[text_block],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    with pytest.raises(Exception) as exc_info:
        extract_and_parse_anthropic_json_response(response)
    assert "No valid AgentResponse format found" in str(exc_info.value)

def test_empty_content():
    # Create a mock response with empty content
    response = Message(
        id="test_id",
        role="assistant",
        content=[],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    with pytest.raises(Exception) as exc_info:
        extract_and_parse_anthropic_json_response(response)
    assert "No valid AgentResponse format found" in str(exc_info.value)

def test_tool_use_without_agent_response():
    # Create a mock response with tool use but not agent_response
    tool_use_block = ToolUseBlock(
        type="tool_use",
        id="test_id",
        name="other_tool",
        input={"key": "value"}
    )
    response = Message(
        id="test_id",
        role="assistant",
        content=[tool_use_block],
        model="claude-3-opus-20240229",
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    
    with pytest.raises(Exception) as exc_info:
        extract_and_parse_anthropic_json_response(response)
    assert "No valid AgentResponse format found" in str(exc_info.value)

CUSTOM_MODEL = "claude-3-5-sonnet-latest"
@pytest.mark.parametrize("response_format_info, expected_model", [
    (ResponseFormatInfo(name="agent_response", format=AgentResponse), CUSTOM_MODEL),  # With response_format_info - should use self.model
    (None, ANTHROPIC_FAST_MODEL),  # Without response_format_info - should use ANTHROPIC_FAST_MODEL
])
@pytest.mark.asyncio
async def test_model_selection_based_on_response_format_info(response_format_info, expected_model):
    """
    Tests that the correct model is selected based on whether response_format_info is provided.
    """
    
    # Create an AnthropicClient with a specific model
    custom_model = CUSTOM_MODEL
    client = AnthropicClient(api_key="test_key", model=custom_model)
    
    # Mock the messages.create method to avoid actual API calls
    client.client = MagicMock()
    mock_response = Message(
        id="test_id",
        role="assistant",
        content=[TextBlock(type="text", text="Test response")],
        model=custom_model,
        type="message",
        usage=Usage(input_tokens=0, output_tokens=0)
    )
    client.client.messages.create.return_value = mock_response
    
    with patch('mito_ai.anthropic_client.get_anthropic_completion_function_params', wraps=get_anthropic_completion_function_params) as mock_get_params:
        await client.request_completions(
            messages=[{"role": "user", "content": "Test message"}],
            response_format_info=response_format_info
        )
        
        # Verify that get_anthropic_completion_function_params was called with the expected model
        mock_get_params.assert_called_once()
        call_args = mock_get_params.call_args
        assert call_args[1]['model'] == expected_model 


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
import pkg_resources
import subprocess
import sys
import uuid
from typing import List, Optional, Tuple, Union
from tornado.httpclient import AsyncHTTPClient

def get_random_id() -> str:
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def is_running_test() -> bool:
    """
    A helper function that quickly returns if the current code is running 
    inside of a test, which is useful for making sure we don't generate 
    tons of logs.
    """
    # Pytest injects PYTEST_CURRENT_TEST into the current environment when running
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    
    # Github injects CI into the environment when running
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci

def get_installed_packages() -> List[str]:
    """
    Get a list of all installed packages.
    """
    return [
        package.key
        for package in sorted(pkg_resources.working_set, key=lambda x: x.key)
    ]

def install_packages(packages: List[str]) -> dict[str, Union[bool, str, None]]:
    """
    Install a list of packages.
    
    Returns:
        dict: A dictionary containing:
            - 'success': bool
            - 'error': error message captured from the subprocess call or None if no error
    """
    result: dict[str, Union[bool, str, None]] = {
        'success': True,
        'error': None
    }
    
    for package in packages:
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        except subprocess.CalledProcessError as e:
            result['success'] = False
            result['error'] = str(e)
    
    return result

def _create_http_client(timeout: int, max_retries: int) -> Tuple[AsyncHTTPClient, Optional[int]]:
    """
    Create an HTTP client with appropriate timeout settings.
    
    Args:
        timeout: The timeout in seconds
        max_retries: The maximum number of retries
        
    Returns:
        A tuple containing the HTTP client and the timeout value in milliseconds
    """
    from .utils import is_running_test  # local import to avoid circular import if needed
    if is_running_test():
        # If we are running in a test environment, setting the request_timeout fails for some reason.
        http_client = AsyncHTTPClient(defaults=dict(user_agent="Mito-AI client"))
        http_client_timeout = None
    else:
        # To avoid 599 client timeout errors, we set the request_timeout. By default, the HTTP client 
        # timesout after 20 seconds. We update this to match the timeout we give to OpenAI. 
        # The OpenAI timeouts are denoted in seconds, whereas the HTTP client expects milliseconds. 
        # We also give the HTTP client a 10 second buffer to account for
        http_client_timeout = timeout * 1000 * max_retries + 10000
        http_client = AsyncHTTPClient(defaults=dict(user_agent="Mito-AI client", request_timeout=http_client_timeout))
    
    return http_client, http_client_timeout



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
from typing import Final
from mito_ai.utils.schema import MITO_FOLDER

SETTINGS_PATH: Final[str] = os.path.join(MITO_FOLDER, "settings.json")
BACKUP_SETTINGS_PATH: Final[str] = os.path.join(MITO_FOLDER, "settings_backup.json") 


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, List, Tuple
from jupyter_server.utils import url_path_join
from mito_ai.log.handlers import LogHandler

def get_log_urls(base_url: str) -> List[Tuple[str, Any, dict]]:
    """Get all log related URL patterns.

    Args:
        base_url: The base URL for the Jupyter server

    Returns:
        List of (url_pattern, handler_class, handler_kwargs) tuples
    """
    BASE_URL = base_url + "/mito-ai"
    
    return [
        (url_path_join(BASE_URL, "log"), LogHandler, {}),
    ]


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, Final, List, Optional
import os
from mito_ai.utils.schema import MITO_FOLDER

RULES_DIR_PATH: Final[str] = os.path.join(MITO_FOLDER, 'rules')

def set_rules_file(rule_name: str, value: Any) -> None:
    """
    Updates the value of a specific rule file in the rules directory
    """
    # Ensure the directory exists
    if not os.path.exists(RULES_DIR_PATH):
        os.makedirs(RULES_DIR_PATH)
    
    # Create the file path to the rule name as a .md file
    file_path = os.path.join(RULES_DIR_PATH, f"{rule_name}.md")
    
    with open(file_path, 'w+') as f:
        f.write(value)
    

def get_rule(rule_name: str) -> Optional[str]:
    """
    Retrieves the value of a specific rule file from the rules directory
    """
    
    if rule_name.endswith('.md'):
        rule_name = rule_name[:-3]
    
    file_path = os.path.join(RULES_DIR_PATH, f"{rule_name}.md")
    
    if not os.path.exists(file_path):
        return None
    
    with open(file_path, 'r') as f:
        return f.read()


def get_all_rules() -> List[str]:
    """
    Retrieves all rule files from the rules directory
    """
    # Ensure the directory exists
    if not os.path.exists(RULES_DIR_PATH):
        os.makedirs(RULES_DIR_PATH)
        return []  # Return empty list if directory didn't exist
    
    try:
        return [f for f in os.listdir(RULES_DIR_PATH) if f.endswith('.md')]
    except OSError as e:
        # Log the error if needed and return empty list
        print(f"Error reading rules directory: {e}")
        return []



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
from datetime import datetime
from .utils import get_random_id

# Current field names
# Some helpful constants
GITHUB_ACTION_ID = 'github_action'
GITHUB_ACTION_EMAIL = 'github@action.com'

# Old field names
UJ_INTENDED_BEHAVIOR = 'intended_behavior'
UJ_CLOSED_FEEDBACK = 'closed_feedback'
UJ_MITOSHEET_LAST_FIVE_USAGES = 'mitosheet_last_five_usages'

# Current field names
UJ_USER_JSON_VERSION = 'user_json_version'
UJ_STATIC_USER_ID = 'static_user_id'
UJ_USER_SALT = 'user_salt'
UJ_USER_EMAIL = 'user_email'
UJ_RECEIVED_TOURS = 'received_tours'
UJ_FEEDBACKS = 'feedbacks'
UJ_FEEDBACKS_V2 = 'feedbacks_v2'
UJ_MITOSHEET_CURRENT_VERSION = 'mitosheet_current_version'
UJ_MITOSHEET_LAST_UPGRADED_DATE = 'mitosheet_last_upgraded_date'
UJ_MITOSHEET_LAST_FIFTY_USAGES = 'mitosheet_last_fifty_usages'
UJ_MITOSHEET_TELEMETRY = 'mitosheet_telemetry'
UJ_MITOSHEET_PRO = 'mitosheet_pro'
UJ_MITOSHEET_ENTERPRISE = 'mitosheet_enterprise'
UJ_EXPERIMENT = 'experiment'
UJ_RECEIVED_CHECKLISTS = 'received_checklists'
UJ_AI_PRIVACY_POLICY = 'ai_privacy_policy'
UJ_AI_MITO_API_NUM_USAGES = 'ai_mito_api_num_usages'
UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES = 'ai_mito_autocomplete_num_usages'
UJ_MITO_AI_FIRST_USAGE_DATE = 'mito_ai_first_usage_date'
UJ_MITO_AI_LAST_RESET_DATE = 'mito_ai_last_reset_date'

MITO_CONFIG_KEY_HOME_FOLDER = 'MITO_CONFIG_HOME_FOLDER'
if MITO_CONFIG_KEY_HOME_FOLDER in os.environ:
    HOME_FOLDER = os.path.expanduser(os.environ[MITO_CONFIG_KEY_HOME_FOLDER])
else:
    HOME_FOLDER = os.path.expanduser('~')

# Where all global .mito files are stored
MITO_FOLDER = os.path.join(HOME_FOLDER, ".mito")

"""
The most up to date version of the user.json object
"""
USER_JSON_VERSION_10 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 10,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: 0,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_MITOSHEET_ENTERPRISE: False,
    UJ_EXPERIMENT: {
        'experiment_id': 'installer_communication_and_time_to_value',
        'variant': 'B',
    },
    UJ_RECEIVED_CHECKLISTS: {},
    UJ_AI_PRIVACY_POLICY: False,
    UJ_AI_MITO_API_NUM_USAGES: 0,
    UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES: 0,
    UJ_MITO_AI_FIRST_USAGE_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITO_AI_LAST_RESET_DATE: datetime.today().strftime('%Y-%m-%d'),
}

# This is the most up to date user json, and you must update it when
# you add a new schema
USER_JSON_DEFAULT = USER_JSON_VERSION_10


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List, Union, AsyncGenerator, Callable
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.completions.models import CodeExplainMetadata, MessageType, CompletionRequest, CompletionStreamChunk, CompletionReply
from mito_ai.completions.prompt_builders.explain_code_prompt import create_explain_code_prompt
from mito_ai.completions.providers import OpenAIProvider
from mito_ai.completions.message_history import GlobalMessageHistory
from mito_ai.completions.completion_handlers.completion_handler import CompletionHandler
from mito_ai.completions.completion_handlers.utils import append_chat_system_message

__all__ = ["get_code_explain_completion", "stream_code_explain_completion"]

class CodeExplainHandler(CompletionHandler[CodeExplainMetadata]):
    """Handler for code explain completions."""
    
    @staticmethod
    async def get_completion(
        metadata: CodeExplainMetadata,
        provider: OpenAIProvider,
        message_history: GlobalMessageHistory,
        model: str
    ) -> str:
        """Get a code explain completion from the AI provider."""
        active_cell_code = metadata.activeCellCode or ''
        thread_id = metadata.threadId

        # Add the system message if it doesn't already exist
        await append_chat_system_message(message_history, model, provider, thread_id)

        # Create the prompt
        prompt = create_explain_code_prompt(active_cell_code)
        display_prompt = f"```python{active_cell_code}```"

        # Add the prompt to the message history
        new_ai_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": prompt}
        new_display_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": display_prompt}
        await message_history.append_message(
            new_ai_optimized_message, new_display_optimized_message, model, provider, thread_id
        )

        # Get the completion
        completion = await provider.request_completions(
            messages=message_history.get_ai_optimized_history(thread_id),
            model=model,
            message_type=MessageType.CODE_EXPLAIN,
            thread_id=thread_id
        )

        # Add the response to message history
        ai_response_message: ChatCompletionMessageParam = {"role": "assistant", "content": completion}
        await message_history.append_message(ai_response_message, ai_response_message, model, provider, thread_id)

        return completion

    @staticmethod
    async def stream_completion(
        metadata: CodeExplainMetadata,
        provider: OpenAIProvider,
        message_history: GlobalMessageHistory,
        message_id: str,
        reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None],
        model: str
    ) -> str:
        """Stream code explain completions from the AI provider.

        Args:
            metadata: The metadata for the code explain completion request.
            provider: The AI provider to use.
            message_history: The message history for this conversation.
            message_id: The ID of the message being processed.
            reply_fn: Function to call with each chunk for streaming replies.

        Returns:
            The accumulated response string.
        """
        active_cell_code = metadata.activeCellCode or ''
        thread_id = metadata.threadId

        # Add the system message if it doesn't already exist
        await append_chat_system_message(message_history, model, provider, thread_id)

        # Create the prompt
        prompt = create_explain_code_prompt(active_cell_code)
        display_prompt = f"```python{active_cell_code}```"

        # Add the prompt to the message history
        new_ai_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": prompt}
        new_display_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": display_prompt}
        await message_history.append_message(
            new_ai_optimized_message, new_display_optimized_message, model, provider, thread_id
        )

        # Stream the completions using the provider's stream method
        accumulated_response = await provider.stream_completions(
            message_type=MessageType.CODE_EXPLAIN,
            messages=message_history.get_ai_optimized_history(thread_id),
            model=model,
            message_id=message_id,
            reply_fn=reply_fn,
            thread_id=thread_id
        )

        # Add the response to message history
        ai_response_message: ChatCompletionMessageParam = {"role": "assistant", "content": accumulated_response}
        await message_history.append_message(ai_response_message, ai_response_message, model, provider, thread_id)

        return accumulated_response

# Use the static methods directly
get_code_explain_completion = CodeExplainHandler.get_completion
stream_code_explain_completion = CodeExplainHandler.stream_completion



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
import pytest
import shutil
from mito_ai.tests.settings.test_settings_constants import (
    SETTINGS_PATH,
    BACKUP_SETTINGS_PATH,
)


@pytest.fixture(scope="module", autouse=True)
def backup_settings_json():
    """Backup the settings.json file before tests and restore it after."""
    if os.path.exists(SETTINGS_PATH):
        # Create backup, so we can test with a clean settings.json file
        shutil.move(SETTINGS_PATH, BACKUP_SETTINGS_PATH)

    yield

    # Cleanup after tests
    if os.path.exists(SETTINGS_PATH):
        os.remove(SETTINGS_PATH)
    if os.path.exists(BACKUP_SETTINGS_PATH):
        shutil.move(BACKUP_SETTINGS_PATH, SETTINGS_PATH)



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
from mito_ai.utils.gemini_utils import _prepare_gemini_request_data_and_headers
from mito_ai.completions.models import MessageType

TEST_CONTENTS = [
    {'role': 'system', 'content': 'You are Mito Data Copilot, an AI assistant for Jupyter.'},
    {'role': 'user', 'content': 'Help me complete the following task. print 10'},
    {'role': 'assistant', 'content': 'python\nprint(10)\n\nPrinted the number 10 '},
    {'role': 'user', 'content': 'Update to print 11'}
]


def test_basic_request_preparation():
    """Test basic request preparation with minimal parameters."""
    model = "gemini-pro"
    message_type = MessageType.CHAT

    data, headers = _prepare_gemini_request_data_and_headers(
        model=model,
        contents=TEST_CONTENTS,
        message_type=message_type
    )

    assert headers == {"Content-Type": "application/json"}
    assert data["timeout"] == 30
    assert data["max_retries"] == 1
    assert data["data"]["model"] == model
    assert data["data"]["contents"] == TEST_CONTENTS
    assert data["data"]["message_type"] == message_type.value


def test_request_with_config():
    """Test request preparation with additional config parameters."""
    model = "gemini-pro"
    message_type = MessageType.CHAT
    config = {
        "temperature": 0.7,
        "max_tokens": 100
    }

    data, headers = _prepare_gemini_request_data_and_headers(
        model=model,
        contents=TEST_CONTENTS,
        message_type=message_type,
        config=config
    )

    assert data["data"]["config"] == config


def test_request_with_response_format():
    """Test request preparation with response format info."""

    class TestFormat:
        name = "test_format"
        format = "json"

    model = "gemini-pro"
    message_type = MessageType.CHAT
    response_format_info = TestFormat()

    data, headers = _prepare_gemini_request_data_and_headers(
        model=model,
        contents=TEST_CONTENTS,
        message_type=message_type,
        response_format_info=response_format_info
    )

    expected_format = {
        "name": "test_format",
        "format": "json"
    }
    assert data["data"]["response_format_info"] == '{"name": "test_format", "format": "json"}'


def test_request_with_complex_config():
    """Test request preparation with complex nested config."""
    model = "gemini-pro"
    message_type = MessageType.CHAT
    config = {
        "temperature": 0.7,
        "nested": {
            "key": "value",
            "array": [1, 2, 3]
        }
    }

    data, headers = _prepare_gemini_request_data_and_headers(
        model=model,
        contents=TEST_CONTENTS,
        message_type=message_type,
        config=config
    )

    assert data["data"]["config"] == config



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
from typing import Union

# Claude
CLAUDE_API_KEY = os.environ.get("CLAUDE_API_KEY")

# Gemini
GEMINI_API_KEY = os.environ.get("GEMINI_API_KEY")

# Ollama
OLLAMA_MODEL = os.environ.get("OLLAMA_MODEL")
OLLAMA_BASE_URL = os.environ.get("OLLAMA_BASE_URL", "http://localhost:11434/v1")

# OpenAI
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")

# Azure OpenAI Config 
AZURE_OPENAI_API_KEY = os.environ.get("AZURE_OPENAI_API_KEY")
AZURE_OPENAI_API_VERSION = os.environ.get("AZURE_OPENAI_API_VERSION")
AZURE_OPENAI_ENDPOINT = os.environ.get("AZURE_OPENAI_ENDPOINT")
AZURE_OPENAI_MODEL = os.environ.get("AZURE_OPENAI_MODEL")

def get_model_provider(model: str) -> Union[str, None]:
    """
    Determine the model type based on the model name prefix
    """
    if not model:
        return None

    model_lower = model.lower()

    if model_lower.startswith('claude'):
        return 'claude'
    elif model_lower.startswith('gemini'):
        return 'gemini'
    elif model_lower.startswith('ollama'):
        return 'ollama'
    elif model_lower.startswith('gpt'):
        return 'openai'

    return None


# Mito AI Base URLs and Endpoint Paths
MITO_PROD_BASE_URL = "https://yxwyadgaznhavqvgnbfuo2k6ca0jboku.lambda-url.us-east-1.on.aws"
MITO_DEV_BASE_URL = "https://x3rafympznv4abp7phos44gzgu0clbui.lambda-url.us-east-1.on.aws"

# Set ACTIVE_BASE_URL manually
ACTIVE_BASE_URL = MITO_PROD_BASE_URL  # Change to MITO_DEV_BASE_URL for dev

# Endpoint paths
ANTHROPIC_PATH = "anthropic/completions"
GEMINI_PATH = "gemini/completions"
OPENAI_PATH = "openai/completions"

# Full URLs (always use ACTIVE_BASE_URL)
MITO_ANTHROPIC_URL = f"{ACTIVE_BASE_URL}/{ANTHROPIC_PATH}"
MITO_GEMINI_URL = f"{ACTIVE_BASE_URL}/{GEMINI_PATH}"
MITO_OPENAI_URL = f"{ACTIVE_BASE_URL}/{OPENAI_PATH}"


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
import threading
import time
from jupyter_server.serverapp import ServerApp
from traitlets.config import Config

TOKEN = "test-token"

@pytest.fixture
def jp_server_config(token_fixture):
    """Configure the Jupyter server for testing."""
    config = Config()
    config.ServerApp.jpserver_extensions = {"mito_ai": True}
    # Disable password requirement for testing
    config.ServerApp.password = ""
    # Set the token for testing
    config.ServerApp.token = token_fixture
    # Enable authentication
    config.ServerApp.allow_unauthenticated_access = False
    return config

@pytest.fixture
def jp_serverapp(jp_server_config, tmp_path):
    """Create a Jupyter server instance for testing."""
    app = ServerApp(config=jp_server_config)
    app.root_dir = str(tmp_path)
    app.initialize(argv=[])

    # Start the server in a non-blocking way
    def start_server():
        app.start()

    server_thread = threading.Thread(target=start_server)
    server_thread.daemon = True  # This ensures the thread will be killed when the main program exits
    server_thread.start()

    # Give the server a moment to start
    time.sleep(1)

    yield app
    app.stop()

@pytest.fixture
def jp_base_url(jp_serverapp):
    """Get the base URL of the Jupyter server."""
    return jp_serverapp.connection_url 

@pytest.fixture
def token_fixture():
    return TOKEN



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

# # Copyright (c) Saga Inc.
# # Distributed under the terms of the GNU Affero General Public License v3.0 License.

# """
# This test will not run in CI, but you can run it locally with:
# python -m pytest mito_ai/tests/performance_test.py -v -s

# Note that you will also need to edit open_ai_utils.py,
# specifically the `is_running_test` condition. I would recommend
# copying the code from the else condition into the if condition.
# """

# import time
# import os
# import pytest
# import openai
# from typing import Generator, List, Dict, Any, Tuple, Optional, cast

# from openai.types.chat import ChatCompletionMessageParam
# from mito_ai.providers import OpenAIProvider
# from mito_ai.completions.models import MessageType
# from mito_ai.tests.data.prompt_sm import prompt_sm
# from mito_ai.tests.data.prompt_lg import prompt_lg
# from mito_ai.tests.data.prompt_xl import prompt_xl

# TEST_MODEL = "o3-mini"
# NUM_ITERATIONS = 10  # Number of requests to make for each test

# # Environment variable to control whether performance tests run in CI
# IS_CI = os.environ.get("CI", "false").lower() == "true"

# # Test messages for all performance tests
# SMALL_TEST_MESSAGE: List[ChatCompletionMessageParam] = [
#     {"role": "user", "content": prompt_sm}
# ]
# LARGE_TEST_MESSAGE: List[ChatCompletionMessageParam] = [
#     cast(
#         ChatCompletionMessageParam,
#         {"role": prompt_lg[i]["role"], "content": prompt_lg[i]["content"]},
#     )
#     for i in range(len(prompt_lg))
# ]
# XL_TEST_MESSAGE: List[ChatCompletionMessageParam] = [
#     cast(
#         ChatCompletionMessageParam,
#         {"role": prompt_xl[i]["role"], "content": prompt_xl[i]["content"]},
#     )
#     for i in range(len(prompt_xl))
# ]

# # Dictionary to store performance metrics for all tests
# ALL_METRICS: Dict[str, Dict[str, Any]] = {}


# async def run_llm_requests(
#     llm: OpenAIProvider,
#     messages: List[ChatCompletionMessageParam],
#     n: int = NUM_ITERATIONS,
# ) -> Tuple[List[Optional[str]], Dict[str, Any]]:
#     """
#     Run LLM requests n times and collect performance metrics.

#     Args:
#         llm: The OpenAIProvider instance to use
#         messages: The messages to send to the LLM
#         n: Number of requests to make (default: NUM_ITERATIONS)

#     Returns:
#         Tuple containing:
#         - List of completion responses
#         - Dictionary with performance metrics (min, max, avg latency)
#     """
#     completions: List[Optional[str]] = []
#     latencies: List[int] = []
#     errors: List[str] = []

#     for i in range(n):
#         try:
#             start_time = time.time()
#             completion = await llm.request_completions(
#                 message_type=MessageType.CHAT, messages=messages, model=TEST_MODEL, thread_id=None
#             )
#             end_time = time.time()

#             latency_ms = round((end_time - start_time) * 1000)
#             latencies.append(latency_ms)
#             completions.append(completion)

#             print(f"Request {i+1}/{n} latency: {latency_ms} ms")
#         except openai.APIError as e:
#             error_msg = f"Request {i+1}/{n} failed: {str(e)}"
#             errors.append(str(e))
#             print(error_msg)
#             completions.append(None)  # Add None for failed completions
#         except Exception as e:
#             error_msg = f"Request {i+1}/{n} failed: {str(e)}"
#             errors.append(str(e))
#             print(error_msg)
#             completions.append(None)  # Add None for failed completions

#     # Calculate stats on successful requests
#     successful_requests = len(latencies)
#     failed_requests = len(errors)

#     metrics: Dict[str, Any] = {
#         "total_requests": n,
#         "successful_requests": successful_requests,
#         "failed_requests": failed_requests,
#     }

#     if successful_requests > 0:
#         avg_latency = sum(latencies) / successful_requests
#         metrics["avg_latency_ms"] = round(avg_latency)
#         metrics["min_latency_ms"] = min(latencies) if latencies else None
#         metrics["max_latency_ms"] = max(latencies) if latencies else None
#         metrics["all_latencies_ms"] = latencies
#     else:
#         metrics["avg_latency_ms"] = None
#         metrics["min_latency_ms"] = None
#         metrics["max_latency_ms"] = None
#         metrics["all_latencies_ms"] = []

#     return completions, metrics


# async def run_direct_openai_requests(
#     client: openai.OpenAI,
#     messages: List[ChatCompletionMessageParam],
#     n: int = NUM_ITERATIONS,
# ) -> Tuple[List[Optional[str]], Dict[str, Any]]:
#     """
#     Run direct OpenAI requests n times and collect performance metrics.

#     Args:
#         client: The OpenAI client instance
#         messages: The messages to send to the API
#         n: Number of requests to make (default: NUM_ITERATIONS)

#     Returns:
#         Tuple containing:
#         - List of completion responses
#         - Dictionary with performance metrics (min, max, avg latency)
#     """
#     completions: List[Optional[str]] = []
#     latencies: List[int] = []
#     errors: List[str] = []

#     for i in range(n):
#         try:
#             start_time = time.time()
#             response = client.chat.completions.create(
#                 model=TEST_MODEL, messages=messages
#             )
#             completion = response.choices[0].message.content
#             end_time = time.time()

#             latency_ms = round((end_time - start_time) * 1000)
#             latencies.append(latency_ms)
#             completions.append(completion)

#             print(f"Direct OpenAI request {i+1}/{n} latency: {latency_ms} ms")
#         except openai.APIError as e:
#             error_msg = f"Direct OpenAI request {i+1}/{n} failed: {str(e)}"
#             errors.append(str(e))
#             print(error_msg)
#             completions.append(None)  # Add None for failed completions
#         except Exception as e:
#             error_msg = f"Direct OpenAI request {i+1}/{n} failed: {str(e)}"
#             errors.append(str(e))
#             print(error_msg)
#             completions.append(None)  # Add None for failed completions

#     # Calculate stats on successful requests
#     successful_requests = len(latencies)
#     failed_requests = len(errors)

#     metrics: Dict[str, Any] = {
#         "total_requests": n,
#         "successful_requests": successful_requests,
#         "failed_requests": failed_requests,
#     }

#     if successful_requests > 0:
#         avg_latency = sum(latencies) / successful_requests
#         metrics["avg_latency_ms"] = round(avg_latency)
#         metrics["min_latency_ms"] = min(latencies) if latencies else None
#         metrics["max_latency_ms"] = max(latencies) if latencies else None
#         metrics["all_latencies_ms"] = latencies
#     else:
#         metrics["avg_latency_ms"] = None
#         metrics["min_latency_ms"] = None
#         metrics["max_latency_ms"] = None
#         metrics["all_latencies_ms"] = []

#     return completions, metrics


# # This fixture runs after all tests and prints the metrics summary
# @pytest.fixture(scope="session", autouse=True)
# def print_metrics_summary() -> Generator[None, Any, None]:
#     """Print a summary of all collected metrics after all tests have run."""
#     yield  # This ensures the code below runs after all tests

#     if ALL_METRICS:
#         print("\n\n" + "=" * 100)
#         print("PERFORMANCE TEST RESULTS SUMMARY")
#         print("=" * 100)

#         # Print in a table format
#         headers = [
#             "Test",
#             "Success/Total",
#             "Success %",
#             "Avg Latency (ms)",
#             "Min Latency (ms)",
#             "Max Latency (ms)",
#         ]
#         row_format = "{:<25} {:<15} {:<10} {:<18} {:<18} {:<18}"

#         print(row_format.format(*headers))
#         print("-" * 100)

#         for test_name, metrics in ALL_METRICS.items():
#             success_rate = "N/A"
#             if "total_requests" in metrics and metrics["total_requests"] > 0:
#                 success_rate = f"{(metrics.get('successful_requests', 0) / metrics['total_requests']) * 100:.1f}%"

#             success_total = f"{metrics.get('successful_requests', 'N/A')}/{metrics.get('total_requests', 'N/A')}"

#             row = [
#                 test_name,
#                 success_total,
#                 success_rate,
#                 metrics.get("avg_latency_ms", "N/A"),
#                 metrics.get("min_latency_ms", "N/A"),
#                 metrics.get("max_latency_ms", "N/A"),
#             ]
#             print(row_format.format(*row))

#         print("=" * 100)


# @pytest.mark.skipif(IS_CI, reason="Performance tests are skipped in CI environments")
# @pytest.mark.asyncio
# async def test_server_key_performance() -> None:
#     """Test the performance of the OpenAI provider when using the server key."""
#     # Save the original API key if it exists
#     original_api_key = os.environ.get("OPENAI_API_KEY", "")

#     try:
#         # Ensure we're using the server key by clearing the API key in environment
#         os.environ["OPENAI_API_KEY"] = ""

#         # Initialize the provider
#         llm = OpenAIProvider()

#         print("\nRunning small prompt")
#         completions_sm, metrics_sm = await run_llm_requests(llm, SMALL_TEST_MESSAGE)
#         ALL_METRICS["Server Key (sm prompt)"] = metrics_sm

#         print("\nRunning large prompt")
#         completions_lg, metrics_lg = await run_llm_requests(llm, LARGE_TEST_MESSAGE)
#         ALL_METRICS["Server Key (lg prompt)"] = metrics_lg

#         print("\nRunning xl prompt")
#         completions_xl, metrics_xl = await run_llm_requests(llm, XL_TEST_MESSAGE)
#         ALL_METRICS["Server Key (xl prompt)"] = metrics_xl
#     finally:
#         # Restore the original API key
#         os.environ["OPENAI_API_KEY"] = original_api_key


# @pytest.mark.skipif(IS_CI, reason="Performance tests are skipped in CI environments")
# @pytest.mark.asyncio
# async def test_user_key_performance() -> None:
#     """Test the performance of the OpenAI provider when using a user key."""
#     # Skip test if no API key is available
#     api_key = os.environ.get("OPENAI_API_KEY")
#     if not api_key:
#         pytest.skip("No OpenAI API key available in environment variables")

#     # Initialize the provider (will use the API key from environment)
#     llm = OpenAIProvider()

#     print("\nRunning sm prompt")
#     completions_sm, metrics_sm = await run_llm_requests(llm, SMALL_TEST_MESSAGE)
#     ALL_METRICS["User Key (sm prompt)"] = metrics_sm

#     print("\nRunning lg prompt")
#     completions_lg, metrics_lg = await run_llm_requests(llm, LARGE_TEST_MESSAGE)
#     ALL_METRICS["User Key (lg prompt)"] = metrics_lg

#     print("\nRunning xl prompt")
#     completions_xl, metrics_xl = await run_llm_requests(llm, XL_TEST_MESSAGE)
#     ALL_METRICS["User Key (xl prompt)"] = metrics_xl


# @pytest.mark.skipif(IS_CI, reason="Performance tests are skipped in CI environments")
# @pytest.mark.asyncio
# async def test_direct_openai_performance() -> None:
#     """Test the performance of direct OpenAI API calls (control group)."""
#     # Skip test if no API key is available
#     api_key = os.environ.get("OPENAI_API_KEY")
#     if not api_key:
#         pytest.skip("No OpenAI API key available in environment variables")

#     # Initialize the OpenAI client directly
#     client = openai.OpenAI(api_key=api_key)

#     print("\nRunning sm prompt")
#     completions_sm, metrics_sm = await run_direct_openai_requests(
#         client, SMALL_TEST_MESSAGE
#     )
#     ALL_METRICS["Direct OpenAI (sm prompt)"] = metrics_sm

#     print("\nRunning lg prompt")
#     completions_lg, metrics_lg = await run_direct_openai_requests(
#         client, LARGE_TEST_MESSAGE
#     )
#     ALL_METRICS["Direct OpenAI (lg prompt)"] = metrics_lg

#     print("\nRunning xl prompt")
#     completions_xl, metrics_xl = await run_direct_openai_requests(
#         client, XL_TEST_MESSAGE
#     )
#     ALL_METRICS["Direct OpenAI (xl prompt)"] = metrics_xl



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import asyncio
import json
import time
import anthropic
from typing import Any, Dict, List, Optional, Union, AsyncGenerator, Tuple, Callable, cast

from anthropic.types import MessageParam, Message, TextBlock, ToolUnionParam
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.completions.models import AgentResponse, MessageType, ResponseFormatInfo, CompletionReply, CompletionStreamChunk, CompletionItem
from mito_ai.utils.schema import UJ_STATIC_USER_ID, UJ_USER_EMAIL
from mito_ai.utils.db import get_user_field
from mito_ai.utils.utils import is_running_test
from mito_ai.utils.server_limits import check_mito_server_quota
from .utils import _create_http_client
from tornado.httpclient import AsyncHTTPClient
from mito_ai.constants import MITO_ANTHROPIC_URL

__user_email: Optional[str] = None
__user_id: Optional[str] = None

timeout = 30
max_retries = 1
INLINE_COMPLETION_MODEL = "claude-3-5-haiku-latest"

def _prepare_anthropic_request_data_and_headers(
    model: Union[str, None],
    max_tokens: int,
    temperature: float,
    system: Union[str, anthropic.NotGiven],
    messages: List[MessageParam],
    message_type: MessageType,
    tools: Optional[List[ToolUnionParam]],
    tool_choice: Optional[dict],
    stream: Optional[bool]
) -> Tuple[Dict[str, Any], Dict[str, str]]:
    check_mito_server_quota(message_type)
    global __user_email, __user_id
    if __user_email is None:
        __user_email = get_user_field(UJ_USER_EMAIL)
    if __user_id is None:
        __user_id = get_user_field(UJ_STATIC_USER_ID)
    # Build the inner data dict (excluding timeout, max_retries, email, user_id)
    inner_data: Dict[str, Any] = {
        "model": model,
        "max_tokens": max_tokens,
        "temperature": temperature,
        "messages": messages
    }
    # Add system to inner_data only if it is not anthropic.NotGiven
    if not isinstance(system, anthropic.NotGiven):
        inner_data["system"] = system
    if tools:
        inner_data["tools"] = tools
    if tool_choice:
        inner_data["tool_choice"] = tool_choice
    if stream:
        inner_data["stream"] = stream
    # Compose the outer data dict
    data = {
        "timeout": timeout,
        "max_retries": max_retries,
        "email": __user_email,
        "user_id": __user_id,
        "data": inner_data
    }
    headers = {"Content-Type": "application/json"}
    return data, headers

async def get_anthropic_completion_from_mito_server(
    model: Union[str, None],
    max_tokens: int,
    temperature: float,
    system: Union[str, anthropic.NotGiven],
    messages: List[MessageParam],
    tools: Optional[List[ToolUnionParam]],
    tool_choice: Optional[dict],
    message_type: MessageType
) -> Message:
    data, headers = _prepare_anthropic_request_data_and_headers(
        model, max_tokens, temperature, system, messages, message_type, tools, tool_choice, None
    )
    http_client, http_client_timeout = _create_http_client(timeout, max_retries)
    start_time = time.time()
    try:
        res = await http_client.fetch(
            MITO_ANTHROPIC_URL,
            method="POST",
            headers=headers,
            body=json.dumps(data),
            request_timeout=http_client_timeout
        )
        print(f"Anthropic request completed in {time.time() - start_time:.2f} seconds")
    except Exception as e:
        print(f"Anthropic request failed after {time.time() - start_time:.2f} seconds with error: {str(e)}")
        raise
    finally:
        http_client.close()
    content = json.loads(res.body)
    # If the response is wrapped in a 'data' field, extract it
    if isinstance(content, dict) and "data" in content:
        return cast(Message, content["data"])
    return cast(Message, content)

async def stream_anthropic_completion_from_mito_server(
    model: Union[str, None],
    max_tokens: int,
    temperature: float,
    system: Union[str, anthropic.NotGiven],
    messages: List[MessageParam],
    stream: bool,
    message_type: MessageType,
    reply_fn: Optional[Callable[[Union[CompletionReply, CompletionStreamChunk]], None]] = None,
    message_id: Optional[str] = None,
) -> AsyncGenerator[str, None]:
    data, headers = _prepare_anthropic_request_data_and_headers(
        model, max_tokens, temperature, system, messages, message_type, None, None, stream
    )
    http_client, http_client_timeout = _create_http_client(timeout, max_retries)
    start_time = time.time()
    chunk_queue: asyncio.Queue[str] = asyncio.Queue()
    fetch_complete = False
    def chunk_callback(chunk: bytes) -> None:
        try:
            chunk_str = chunk.decode('utf-8')
            asyncio.create_task(chunk_queue.put(chunk_str))
        except Exception as e:
            print(f"Error processing Anthropic streaming chunk: {str(e)}")
    fetch_future = None
    try:
        fetch_future = http_client.fetch(
            MITO_ANTHROPIC_URL,
            method="POST",
            headers=headers,
            body=json.dumps(data),
            request_timeout=http_client_timeout,
            streaming_callback=chunk_callback
        )
        async def wait_for_fetch() -> None:
            try:
                await fetch_future
                nonlocal fetch_complete
                fetch_complete = True
                print("Anthropic fetch completed")
            except Exception as e:
                print(f"Error in Anthropic fetch: {str(e)}")
                raise
        fetch_task = asyncio.create_task(wait_for_fetch())
        while not (fetch_complete and chunk_queue.empty()):
            try:
                chunk = await asyncio.wait_for(chunk_queue.get(), timeout=0.1)
                if reply_fn and message_id:
                    reply_fn(CompletionStreamChunk(
                        parent_id=message_id,
                        chunk=CompletionItem(
                            content=chunk,
                            isIncomplete=True,
                            token=message_id,
                        ),
                        done=False,
                    ))
                yield chunk
            except asyncio.TimeoutError:
                if fetch_complete and chunk_queue.empty():
                    break
                continue
        print(f"\nAnthropic stream completed in {time.time() - start_time:.2f} seconds")
        if reply_fn and message_id:
            reply_fn(CompletionStreamChunk(
                parent_id=message_id,
                chunk=CompletionItem(
                    content="",
                    isIncomplete=False,
                    token=message_id,
                ),
                done=True,
            ))
    except Exception as e:
        print(f"\nAnthropic stream failed after {time.time() - start_time:.2f} seconds with error: {str(e)}")
        if fetch_future:
            try:
                await fetch_future
            except Exception:
                pass
        raise
    finally:
        http_client.close()

def get_anthropic_completion_function_params(
    model: str,
    messages: List[MessageParam],
    max_tokens: int,
    system: Union[str, anthropic.NotGiven],
    temperature: float = 0.0,
    tools: Optional[List[ToolUnionParam]] = None,
    tool_choice: Optional[dict] = None,
    stream: Optional[bool] = None,
    response_format_info: Optional[ResponseFormatInfo] = None,
) -> Dict[str, Any]:
    """
    Build the provider_data dict for Anthropic completions, mirroring the OpenAI approach.
    Only includes fields needed for the Anthropic API.
    """
    provider_data = {
        "model": model,
        "max_tokens": max_tokens,
        "temperature": temperature,
        "messages": messages,
        "system": system,
    }
    if response_format_info is not None:
        provider_data["model"] = INLINE_COMPLETION_MODEL
    if tools:
        provider_data["tools"] = tools
    if response_format_info and response_format_info.name == "agent_response":
        provider_data["tools"] = [{
            "name": "agent_response",
            "description": "Output a structured response following the AgentResponse format",
            "input_schema": AgentResponse.model_json_schema()
        }]
        provider_data["tool_choice"] = {"type": "tool", "name": "agent_response"}
    
    
    if tool_choice:
        provider_data["tool_choice"] = tool_choice
    if stream is not None:
        provider_data["stream"] = stream
    # Optionally handle response_format_info if Anthropic supports it in the future
    return provider_data



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mito_ai.completions.prompt_builders.prompt_constants import (
    CITATION_RULES,
    FILES_SECTION_HEADING,
    JUPYTER_NOTEBOOK_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    get_database_rules
)


def create_agent_system_message_prompt(isChromeBrowser: bool) -> str:
    
    # The GET_CELL_OUTPUT tool only works on Chrome based browsers. 
    # This constant helps us replace the phrase 'or GET_CELL_OUTPUT' with ''
    # throughout the prompt
    OR_GET_CELL_OUTPUT = 'or GET_CELL_OUTPUT' if isChromeBrowser else ''

    return f"""You are Mito Data Copilot, an AI assistant for Jupyter. You're a great python programmer, a seasoned data scientist and a subject matter expert.

The user is going to ask you to guide them as they complete a task. You will help them complete a task over the course of an entire conversation with them. The user will first share with you what they want to accomplish. You will then give them the first step of the task, they will apply that first step, share the updated notebook state with you, and then you will give them the next step of the task. You will continue to give them the next step of the task until they have completed the task.

You have access to a set of tools that you can use to accomplish the task you've been given. You can use one tool per message, and will receive the result of that tool use in the user's response. You use tools step-by-step to accomplish a given task, with each tool use informed by the result of the previous tool use.

Each time you use a tool, except for the finished_task tool, the user will execute the tool and provide you with updated information about the notebook and variables defined in the kernel to help you decide what to do next.

====

TOOL: CELL_UPDATE

CELL_UPDATE is how you communicate to the user about the changes you want to make to the notebook. Each CELL_UPDATE can either modify an existing cell or create a new cell. 

There are two types of CELL_UPDATE:

1. CellModification
2. CellAddition

Each time you want to make a change to the notebook, you will respond with a CellModification or CellAddition.

#### Cell Modification
When you want to modify an existing cell in the notebook, respond in this format.

Format:
{{
    type: 'cell_update',
    message: str,
    cell_update: {{
        type: 'modification'
        id: str,
        code: str
        cell_type: 'code' | 'markdown'
    }}
    get_cell_output_cell_id: None,
    next_steps: None
}}

Important information:
1. The id is the id of the code cell that you want to update. The id MUST already be part of the original Jupyter Notebook that your colleague shared with you.
2. The message is a short summary of your thought process that helped you decide what to update in cell_update.
3. The code should be the full contents of that updated code cell. The code that you return will overwrite the existing contents of the code cell so it must contain all necessary code.
4. Important: Only use the CELL_UPDATE tool if you want to add/modify a notebook cell in response to the user's request. If the user is just sending you a friendly greeting or asking you a question about yourself, you SHOULD NOT USE A CELL_UPDATE tool because it does not require modifying the notebook. Instead, just use the FINISHED_TASK response.

#### Cell Addition:
When you want to add a new cell to the notebook, respond in this format

Format: 
{{
    type: 'cell_update',
    message: str,
    cell_update: {{
        type: 'new'
        index: int
        code: str   
        cell_type: 'code' | 'markdown'
    }}
    get_cell_output_cell_id: None,
    next_steps: None
}}

Important information:
1. The index should be the 0-index position of where you want the new code cell to be added in the notebook.
2. The message is a short summary of your thought process that helped you decide what to update in cell_update.
3. The code should be the full contents of that updated code cell. The code that you return will overwrite the existing contents of the code cell so it must contain all necessary code.
4. The cell_type should only be 'markdown' if there is no code to add. There may be times where the code has comments. These are still code cells and should have the cell_type 'code'. Any cells that are labeled 'markdown' will be converted to markdown cells by the user.

<Cell Modification Example>
Jupyter Notebook:
[
    {{
        cell_type: 'markdown'
        id: '9e38c62b-38f8-457d-bb8d-28bfc52edf2c'
        code: \"\"\" # Used Car Sales Analysis \"\"\"
    }},
    {{
        cell_type: 'code'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: \"\"\"import pandas as pd
sales_df = pd.read_csv('./sales.csv') 
loan_multiplier = 1.5\"\"\"
    }},
]

{VARIABLES_SECTION_HEADING}
{{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

{FILES_SECTION_HEADING}
file_name: sales.csv

Your task: 
Convert the transaction_date column to datetime and then multiply the total_price column by the sales_multiplier.

Output:
{{
    type: 'cell_update',
    cell_type: 'code',
    cell_update: {{
        type: 'modification'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc',
        code: "import pandas as pd\\nsales_df = pd.read_csv('./sales.csv')\\nloan_multiplier = 1.5\\nsales_df['transaction_date'] = pd.to_datetime(sales_df['transaction_date'])\\nsales_df['total_price'] = sales_df['total_price'] * sales_multiplier",
        cell_type: 'code'
    }},
    get_cell_output_cell_id: None,
    next_steps: None
}}

</Cell Modification Example>

<Cell Addition Example>
{JUPYTER_NOTEBOOK_SECTION_HEADING}
[
    {{
        cell_type: 'markdown'
        id: '9e38c62b-38f8-457d-bb8d-28bfc52edf2c'
        code: \"\"\"# Used Car Sales Analysis \"\"\"
    }},
    {{
        cell_type: 'code'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: \"\"\"import pandas as pd
sales_df = pd.read_csv('./sales.csv')
sales_df['transaction_date'] = pd.to_datetime(sales_df['transaction_date'])\"\"\"
    }},
]

{VARIABLES_SECTION_HEADING}
{{
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

{FILES_SECTION_HEADING}
file_name: sales.csv

Your task: 
Graph the total_price for each sale

Output:
{{
    type: 'cell_update',
    message: "I'll create a graph with using matplotlib with sale `index` on the x axis and `total_price` on the y axis.",
    cell_update: {{
        type: 'add'
        index: 2
        code: "import matplotlib.pyplot as plt\n\nplt.bar(sales_df.index, sales_df['total_price'])\nplt.title('Total Price per Sale')\nplt.xlabel('Transaction Number')\nplt.ylabel('Sales Price ($)')\nplt.show()"
    }},
    get_cell_output_cell_id: None,
    next_steps: None
}}

</Cell Addition Example>

{'' if not isChromeBrowser else '''====

TOOL: GET_CELL_OUTPUT

When you want to get a base64 encoded version of a cell's output, respond with this format:

{{
    type: 'get_cell_output',
    message: str,
    get_cell_output_cell_id: str,
    cell_update: None,
    next_steps: Optional[List[str]]
}}

Important information:
1. The message is a short summary of the description of why you want to get the cell output. For example: "Let's check the graph to make sure it's readable"
2. The cell_id is the id of the cell that you want to get the output from.

===='''
}

TOOL: FINISHED_TASK

When you have completed the user's task, respond with a message in this format:

{{
    type: 'finished_task',
    message: str,
    get_cell_output_cell_id: None,
    cell_update: None,
    next_steps: Optional[List[str]]
}}

Important information:
1. The message is a short summary of the ALL the work that you've completed on this task. It should not just refer to the final message. It could be something like "I've completed the sales strategy analysis by exploring key relationships in the data and summarizing creating a report with three recommendations to boost sales.""
2. The message should include citations for any insights that you shared with the user.
3. The next_steps is an optional list of 2 or 3 suggested follow-up tasks or analyses that the user might want to perform next. These should be concise, actionable suggestions that build on the work you've just completed. For example: ["Visualize the results", "Export the cleaned data to CSV", "Perform statistical analysis on the key metrics"].
4. The next_steps should be as relevant to the user's actual task as possible. Try your best not to make generic suggestions like "Analyze the data" or "Visualize the results". For example, if the user just asked you to calculate LTV of their customers, you might suggest the following next steps: ["Graph key LTV drivers: churn and average transaction value", "Visualize LTV per customer age group"].
5. If you are not sure what the user might want to do next, err on the side of suggesting next steps instead of making an assumption and using more CELL_UPDATES.
6. If the user's task doesn't involve creating or modifying a code cell, you should respond with a FINISHED_TASK response. 
7. If the user is just sending a friendly greeting (like "Hello", "Hi", "Hey", "How are you?", "What can you help me with?", etc.), you must respond with a FINISHED_TASK response message with a friendly message like this: "Hello! I'm Mito AI, your AI assistant for data analysis and Python programming in Jupyter notebooks. I can help you analyze datasets, create visualizations, clean data, and much more. What would you like to work on today?"

<Finished Task Example 1>

{{
    type: 'finished_task',
    message: "Revenue analysis complete: total sales reached $2.3M with 34% growth in Q4[MITO_CITATION:abc123:2-3], while premium products generated 67% of profit margins[MITO_CITATION:xyz456:5]. The customer segmentation workflow identified three distinct buying patterns driving conversion rates[MITO_CITATION:def456:8-12].",
    get_cell_output_cell_id: None,
    cell_update: None,
    next_steps: ["Graph sales by product category", "Identify seasonal patterns in data", "Find the top 3 performing products"]
}}

</Finished Task Example 1>

<Finished Task Example 2>

User message: "Hi"

Output:
{{
    type: 'finished_task',
    message: "Hey there! I'm Mito AI. How can I help you today?",
    get_cell_output_cell_id: None,
    cell_update: None,
    next_steps: None
}}

</Finished Task Example 2>

====

RULES

- You are working in a Jupyter Lab environment in a .ipynb file. 
- You can only respond with CELL_UPDATES or FINISHED_TASK responses.
- In each message you send to the user, you can send one CellModification, one CellAddition, or one FINISHED_TASK response. BUT YOU WILL GET TO SEND MULTIPLE MESSAGES TO THE USER TO ACCOMPLISH YOUR TASK SO DO NOT TRY TO ACCOMPLISH YOUR TASK IN A SINGLE MESSAGE.
- After you send a CELL_UPDATE, the user will send you a message with the updated variables, code, and files in the current directory. You will use this information to decide what to do next, so it is critical that you wait for the user's response after each CELL_UPDATE before deciding your next action.
- When updating code, keep as much of the original code as possible and do not recreate variables that already exist.
- When you want to display a dataframe to the user, just write the dataframe on the last line of the code cell instead of writing print(<dataframe name>). Jupyter will automatically display the dataframe in the notebook.
- When writing the message, do not explain to the user how to use the CELL_UPDATE or FINISHED_TASK response, they will already know how to use them. Just provide a summary of your thought process. Do not reference any Cell IDs in the message.
- When writing the message, do not include leading words like "Explanation:" or "Thought process:". Just provide a summary of your thought process.
- When writing the message, use tickmarks when referencing specific variable names. For example, write `sales_df` instead of "sales_df" or just sales_df.

==== 
{CITATION_RULES}

<Citation Example>

### User Message 1:

{JUPYTER_NOTEBOOK_SECTION_HEADING}
[
    {{
        cell_type: 'markdown'
        id: '9e38c62b-38f8-457d-bb8d-28bfc52edf2c'
        code: \"\"\" # Used Car Sales Analysis \"\"\"
    }},
    {{
        cell_type: 'code'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: \"\"\"import pandas as pd
tesla_stock_prices_df = pd.read_csv('./tesla_stock_prices.csv)\"\"\"
    }}
]

{VARIABLES_SECTION_HEADING}
{{
    'tesla_stock_prices_df': pd.DataFrame({{
        'Date': ['2025-01-02', '2024-01-03', '2024-01-04', '2024-01-05', '2024-01-06'],
        'closing_price': [249.98, 251.03, 250.11, 249.97, 251.45]
    }})
}}

{FILES_SECTION_HEADING}
file_name: tesla_stock_prices.csv

Your task: 
Given the dataframe `tesla_stock_prices_df`, what day was Tesla's all time high closing price?

Output:
{{
    type: 'cell_update',
    message: "I'll calculate two new variables all_time_high_date and all_time_high_price.",
    cell_update: {{
        type: 'add'
        index: 2
        code: "all_time_high_row_idx = tesla_stock_prices_df['closing_price'].idxmax()\nall_time_high_date = tesla_stock_prices_df.at[all_time_high_row_idx, 'Date']\nall_time_high_price = tesla_stock_prices_df.at[all_time_high_row_idx, 'closing_price']"
    }},
    get_cell_output_cell_id: None
}}

### User Message 2

{JUPYTER_NOTEBOOK_SECTION_HEADING}
[
    {{
        cell_type: 'markdown'
        id: '9e38c62b-38f8-457d-bb8d-28bfc52edf2c'
        code: \"\"\" # Used Car Sales Analysis \"\"\"
    }},
    {{
        cell_type: 'code'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: \"\"\"import pandas as pd
tesla_stock_prices_df = pd.read_csv('./tesla_stock_prices.csv)\"\"\"
    }},
    {{
        cell_type: 'code',
        id: '9c0d5fda-2b16-4f52-a1c5-a48892f3e2e8',
        code: \"\"\"all_time_high_row_idx = tesla_stock_prices_df['closing_price'].idxmax()
all_time_high_date = tesla_stock_prices_df.at[all_time_high_row_idx, 'Date']
all_time_high_price = tesla_stock_prices_df.at[all_time_high_row_idx, 'closing_price']\"\"\"
    }}
]

{VARIABLES_SECTION_HEADING}
{{
    'tesla_stock_prices_df': pd.DataFrame({{
        'Date': ['2025-01-02', '2024-01-03', '2024-01-04', '2024-01-05', '2024-01-06'],
        'closing_price': [249.98, 251.03, 250.11, 249.97, 251.45],
        'all_time_high_row_idx': 501,
        'all_time_high_date': '2025-03-16',
        'all_time_high_price': 265.91
    }})
}}

{FILES_SECTION_HEADING}
file_name: tesla_stock_prices.csv

Your task: 

Output:
{{
    type: 'finished_task', 
    message: "The all time high tesla stock closing price was $265.91 [MITO_CITATION:9c0d5fda-2b16-4f52-a1c5-a48892f3e2e8:2] on 2025-03-16 [MITO_CITATION:9c0d5fda-2b16-4f52-a1c5-a48892f3e2e8:1]",
    get_cell_output_cell_id: None,
    cell_update: None,
    next_steps: ["Create a visualization of Tesla's stock price over time", "Calculate the percentage change from the lowest to highest price", "Analyze the volatility of Tesla's stock"]
}}

</Cell Addition Example>

===
{get_database_rules()}

====

RULES OF YOUR WORKING PROCESS

The user is going to ask you to guide them as through the process of completing a task. You will help them complete a task over the course of an entire conversation with them. The user will first share with you what they want to accomplish. You will then use a tool to execute the first step of the task, they will execute the tool and return to you the updated notebook state with you, and then you will give them the next step of the task. You will continue to give them the next step of the task until they have completed the task.

As you are guiding the user through the process of completing the task, send them TOOL messages to give them the next step of the task. When you have finished the task, send a FINISHED_TASK tool message. 

The user is a beginning Python user, so you will need to be careful to send them only small steps to complete. Don't try to complete the task in a single response to the user. Instead, each message you send to the user should only contain a single, small step towards the end goal. When the user has completed the step, they will let you know that they are ready for the next step. 

You will keep working in the following iterative format until you have decided that you have finished the user's request. When you decide that you have finished the user's request, respond with a FINISHED_TASK tool message. Otherwise, if you have not finished the user's request, respond with a CELL_UPDATE {OR_GET_CELL_OUTPUT} tool message. When you respond with a CELL_UPDATE, the user will apply the CELL_UPDATE to the notebook and run the new code cell. The user will then send you a message with an updated version of the variables defined in the kernel, code in the notebook, and files in the current directory. In addition, the user will check if the code you provided produced an errored when executed. If it did produce an error, the user will share the error message with you.

Whenever you get a message back from the user, you should:
1. Ask yourself if the previous message you sent to the user was correct. You can answer this question by reviewing the updated code, variables, or output of the cell if you requested it.
2. Ask yourself if you can improve the code or results you got from the previous CELL_UPDATE {OR_GET_CELL_OUTPUT}. If you can, send a new CELL_UPDATE to modify the code you wrote. Improvements might include things like making the code more readable or robust, making sure the code handles reasonable edge cases, improving the output (like making a graph more readable), etc.
3. Decide if you have finished the user's request to you. If you have, respond with a FINISHED_TASK tool message.
4. If you have not finished the user's request, create the next CELL_UPDATE or {OR_GET_CELL_OUTPUT} tool message. 
5. If its not clear what the user want to do next, err on the side of creating a finished_task message with suggested next steps instead of making an assumption and using more CELL_UPDATES. The user might get frustrated if you send irrelevant CELL_UPDATES that do not match their original request.

REMEMBER, YOU ARE GOING TO COMPLETE THE USER'S TASK OVER THE COURSE OF THE ENTIRE CONVERSATION -- YOU WILL GET TO SEND MULTIPLE MESSAGES TO THE USER TO ACCOMPLISH YOUR TASK SO DO NOT TRY TO ACCOMPLISH YOUR TASK IN A SINGLE MESSAGE. IT IS CRUCIAL TO PROCEED STEP-BY-STEP WITH THE SMALLEST POSSIBLE CELL_UPDATES. For example, if asked to build a new dataframe, then analyze it, and then graph the results, you should proceed as follows. 
- Send a CellAddition to add a new code cell to the notebook that creates the dataframe.
- Wait for the user to send you back the updated variables and notebook state so you can decide how to analyze the dataframe.
- Use the data that the user sent you to decide how to analyze the dataframe. Send a CellAddition to add the dataframe analysis code to the notebook.
- Wait for the user to send you back the updated variables and notebook state so you can decide how to proceed. 
- If after reviewing the updates provided by the user, you decide that you want to update the analysis code, send a CellModification to modify the code you just wrote.
- Wait for the user to send you back the updated variables and notebook state so you can decide how to proceed.
- If you are happy with the analysis, refer back to the original task provided by the user to decide your next steps. In this example, it is to graph the results, so you will send a CellAddition to construct the graph. 
- Wait for the user to send you back the updated variables and notebook state.
{'' if not isChromeBrowser else '- Send a GET_CELL_OUTPUT tool message to get the output of the cell you just created and check if you can improve the graph to make it more readable, informative, or professional.'}
- If after reviewing the updates you decide that you've completed the task, send a FINISHED_TASK tool message."""


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
import requests
from mito_ai.tests.db.test_db_constants import SNOWFLAKE
from mito_ai.tests.conftest import TOKEN

# --- GET SCHEMAS ---


def test_get_schemas_with_auth(jp_base_url: str) -> None:
    # Add a connection to get a schema
    response = requests.post(
        jp_base_url + f"/mito-ai/db/connections",
        headers={"Authorization": f"token {TOKEN}"},
        json=SNOWFLAKE,
    )
    assert response.status_code == 200

    # Get the schemas
    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    # Esnure that there is one scema dict in the response
    assert len(response.json()) == 1


def test_ensure_no_application_databases(jp_base_url: str, first_connection_id: str) -> None:
    # We currently don't support Snowflake application databases.
    # This test ensures that we don't crawl application databases.

    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
        headers={"Authorization": f"token {TOKEN}"},
    )

    crawled_databases = response.json()[first_connection_id]["databases"].keys()
    assert len(crawled_databases) > 0
    assert "SNOWFLAKE" not in crawled_databases


def test_get_schemas_with_no_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
    )
    assert response.status_code == 403  # Forbidden


def test_get_schemas_with_incorrect_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden


# --- DELETE SCHEMA ---


def test_delete_schema_with_auth(jp_base_url: str, first_connection_id: str) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/schemas/{first_connection_id}",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200

    # Ensure the schema was deleted
    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    assert len(response.json()) == 0


def test_delete_schema_with_no_auth(jp_base_url: str, first_connection_id: str) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/schemas/{first_connection_id}",
    )
    assert response.status_code == 403  # Forbidden


def test_delete_schema_with_incorrect_auth(
    jp_base_url: str, first_connection_id: str
) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/schemas/{first_connection_id}",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from __future__ import annotations
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional, Union

import openai
from openai.types.chat import ChatCompletionMessageParam
from traitlets import Instance, Unicode, default, validate
from traitlets.config import LoggingConfigurable

from mito_ai import constants
from mito_ai.enterprise.utils import is_azure_openai_configured
from mito_ai.logger import get_logger
from mito_ai.completions.models import (
    AICapabilities,
    CompletionError,
    CompletionItem,
    CompletionItemError,
    CompletionReply,
    CompletionStreamChunk,
    MessageType,
    ResponseFormatInfo,
)
from mito_ai.utils.open_ai_utils import (
    check_mito_server_quota,
    get_ai_completion_from_mito_server,
    get_open_ai_completion_function_params,
    stream_ai_completion_from_mito_server,
)
from mito_ai.utils.server_limits import update_mito_server_quota
from mito_ai.utils.telemetry_utils import (
    MITO_SERVER_KEY,
    USER_KEY,
)

OPENAI_MODEL_FALLBACK = "gpt-4.1"

OPENAI_FAST_MODEL = "gpt-4.1-nano"

class OpenAIClient(LoggingConfigurable):
    """Provide AI feature through OpenAI services."""

    api_key = Unicode(
        config=True,
        allow_none=True,
        help="OpenAI API key. Default value is read from the OPENAI_API_KEY environment variable.",
    )

    last_error = Instance(
        CompletionError,
        allow_none=True,
        help="""Last error encountered when using the OpenAI provider.

This attribute is observed by the websocket provider to push the error to the client.""",
    )

    # Consider the request a failure if it takes longer than 45 seconds.
    # We will try a total of 3 times. Once on the initial request,
    # and then twice more if the first request fails.
    # Note that max_retries cannot be set to None. If we want to disable it, set it to 0.
    timeout = 30
    max_retries = 1

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        super().__init__(log=get_logger(), **kwargs)
        self.last_error = None
        self._async_client: Optional[openai.AsyncOpenAI] = None

    @default("api_key")
    def _api_key_default(self) -> Optional[str]:
        default_key = constants.OPENAI_API_KEY
        return self._validate_api_key(default_key)

    @validate("api_key")
    def _validate_api_key(self, api_key: Optional[str]) -> Optional[str]:
        if not api_key:
            self.log.debug(
                "No OpenAI API key provided; following back to Mito server API."
            )
            return None

        client = openai.OpenAI(api_key=api_key)
        try:
            # Make an http request to OpenAI to make sure it works
            client.models.list()
        except openai.AuthenticationError as e:
            self.log.warning(
                "Invalid OpenAI API key provided.",
                exc_info=e,
            )
            self.last_error = CompletionError.from_exception(
                e,
                hint="You're missing the OPENAI_API_KEY environment variable. Run the following code in your terminal to set the environment variable and then relaunch the jupyter server `export OPENAI_API_KEY=<your-api-key>`",
            )
            return None
        except openai.PermissionDeniedError as e:
            self.log.warning(
                "Invalid OpenAI API key provided.",
                exc_info=e,
            )
            self.last_error = CompletionError.from_exception(e)
            return None
        except openai.InternalServerError as e:
            self.log.debug(
                "Unable to get OpenAI models due to OpenAI error.", exc_info=e
            )
            return api_key
        except openai.RateLimitError as e:
            self.log.debug(
                "Unable to get OpenAI models due to rate limit error.", exc_info=e
            )
            return api_key
        except openai.APIConnectionError as e:
            self.log.warning(
                "Unable to connect to OpenAI API.",
                exec_info=e,
            )
            self.last_error = CompletionError.from_exception(e)
            return None
        else:
            self.log.debug("User OpenAI API key validated.")
            return api_key
        
    @property
    def capabilities(self) -> AICapabilities:
        """Get the provider capabilities."""

        if is_azure_openai_configured():
            return AICapabilities(
                configuration={
                    "model": constants.AZURE_OPENAI_MODEL
                },
                provider="Azure OpenAI",
            )

        if constants.OLLAMA_MODEL and not self.api_key:
            return AICapabilities(
                configuration={
                    "model": constants.OLLAMA_MODEL
                },
                provider="Ollama",
            )

        if self.api_key:
            self._validate_api_key(self.api_key)

            return AICapabilities(
                configuration={
                    "model": OPENAI_MODEL_FALLBACK,
                },
                provider="OpenAI (user key)",
            )

        try:
            check_mito_server_quota(MessageType.CHAT)
        except Exception as e:
            self.log.warning("Failed to set first usage date in user.json", exc_info=e)
            self.last_error = CompletionError.from_exception(e)

        return AICapabilities(
            configuration={
                "model": OPENAI_MODEL_FALLBACK,
            },
            provider="Mito server",
        )

    @property
    def _active_async_client(self) -> Optional[openai.AsyncOpenAI]:
        if not self._async_client or self._async_client.is_closed():
            self._async_client = self._build_openai_client()
        return self._async_client
    
    
    @property
    def key_type(self) -> str:
        """Returns the authentication key type being used."""

        if self.api_key:
            return USER_KEY

        if constants.OLLAMA_MODEL:
            return "ollama"

        return MITO_SERVER_KEY

    def _build_openai_client(self) -> Optional[Union[openai.AsyncOpenAI, openai.AsyncAzureOpenAI]]:
        base_url = None
        llm_api_key = None
        
        if is_azure_openai_configured():
            self.log.debug(f"Using Azure OpenAI with model: {constants.AZURE_OPENAI_MODEL}")
                
            # The format for using Azure OpenAI is different than using
            # other providers, so we have a special case for it here.
            # Create Azure OpenAI client with explicit arguments
            return openai.AsyncAzureOpenAI(
                api_key=constants.AZURE_OPENAI_API_KEY,
                api_version=constants.AZURE_OPENAI_API_VERSION,
                azure_endpoint=constants.AZURE_OPENAI_ENDPOINT or OPENAI_MODEL_FALLBACK,
                max_retries=self.max_retries,
                timeout=self.timeout,
            )
        
        elif constants.OLLAMA_MODEL and not self.api_key:
            base_url = constants.OLLAMA_BASE_URL
            llm_api_key = "ollama"
            self.log.debug(f"Using Ollama with model: {constants.OLLAMA_MODEL}")
        elif self.api_key:
            llm_api_key = self.api_key
            self.log.debug("Using OpenAI with user-provided API key")
        else:
            self.log.warning("No valid API key or model configuration provided")
            return None

        # Create the client with explicit arguments to satisfy type checking
        client = openai.AsyncOpenAI(
            api_key=llm_api_key,
            max_retries=self.max_retries,
            timeout=self.timeout,
            base_url=base_url if base_url else None,
        )
        return client

    def _resolve_model(self, model: Optional[str] = None) -> str:
        if is_azure_openai_configured() and constants.AZURE_OPENAI_MODEL is not None:
            return constants.AZURE_OPENAI_MODEL
        elif constants.OLLAMA_MODEL is not None:
            return constants.OLLAMA_MODEL
        elif model:
            return model
        return OPENAI_MODEL_FALLBACK

    async def request_completions(
            self,
            message_type: MessageType,
            messages: List[ChatCompletionMessageParam],
            model: str,
            response_format_info: Optional[ResponseFormatInfo] = None,
    ) -> str:
        """
        Request completions from the OpenAI API.

        Args:
            message_type: The type of message to request completions for.
            messages: The messages to request completions for.
            model: The model to request completions for.
        Returns:
            The completion from the OpenAI API.
        """
        # Reset the last error
        self.last_error = None
        completion = None

        try:
            model = self._resolve_model(model) if response_format_info else OPENAI_FAST_MODEL

            # Handle other providers as before
            completion_function_params = get_open_ai_completion_function_params(
                model, messages, False, response_format_info
            )

            if self._active_async_client is not None:
                response = await self._active_async_client.chat.completions.create(**completion_function_params)
                completion = response.choices[0].message.content or ""
            else: 
                last_message_content = str(messages[-1].get("content", "")) if messages else None
                completion = await get_ai_completion_from_mito_server(
                    last_message_content,
                    completion_function_params,
                    self.timeout,
                    self.max_retries,
                    message_type,
                )

                update_mito_server_quota(message_type)

            return completion
                
        except BaseException as e:
            self.last_error = CompletionError.from_exception(e)
            raise
    
    async def stream_completions(
        self,
        message_type: MessageType,
        messages: List[ChatCompletionMessageParam],
        model: str,
        message_id: str,
        thread_id: str,
        reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None],
        user_input: Optional[str] = None,
        response_format_info: Optional[ResponseFormatInfo] = None
    ) -> str:
        """
        Stream completions from the OpenAI API and return the accumulated response.
        Returns: The accumulated response string.
        """
        # Reset the last error
        self.last_error = None
        accumulated_response = ""
        
        # Validate that the model is supported.
        model = self._resolve_model(model)
            
        # Send initial acknowledgment
        reply_fn(CompletionReply(
            items=[
                CompletionItem(content="", isIncomplete=True, token=message_id)
            ],
            parent_id=message_id,
        ))

        # Handle other providers as before
        completion_function_params = get_open_ai_completion_function_params(
            model, messages, True, response_format_info
        )

        try:
            if self._active_async_client is not None:
                # Stream from OpenAI
                client = self._active_async_client
                if client is None:
                    raise ValueError("OpenAI client not initialized")
                
                stream = await client.chat.completions.create(**completion_function_params)

                async for chunk in stream:
                    if len(chunk.choices) == 0:
                        continue
                    
                    is_finished = chunk.choices[0].finish_reason is not None
                    content = chunk.choices[0].delta.content or ""
                    accumulated_response += content
                    
                    reply_fn(CompletionStreamChunk(
                        parent_id=message_id,
                        chunk=CompletionItem(
                            content=content,
                            isIncomplete=True,
                            token=message_id,
                        ),
                        done=is_finished,
                    ))
            else:
                # Stream from Mito server
                last_message_content = str(messages[-1].get("content", "")) if messages else ""
                async for chunk_from_mito_server in stream_ai_completion_from_mito_server(
                    last_message_content,
                    completion_function_params,
                    self.timeout,
                    self.max_retries,
                    message_type,
                    reply_fn=reply_fn,
                    message_id=message_id,
                ):
                    accumulated_response += str(chunk_from_mito_server)
                
                # Update quota after streaming is complete
                update_mito_server_quota(message_type)

            return accumulated_response

        except BaseException as e:
            self.last_error = CompletionError.from_exception(e)
            # Send error message to client before raising
            reply_fn(CompletionStreamChunk(
                parent_id=message_id,
                chunk=CompletionItem(
                    content="",
                    isIncomplete=True,
                    error=CompletionItemError(
                        message=f"Failed to process completion: {e!r}"
                    ),
                    token=message_id,
                ),
                done=True,
                error=CompletionError.from_exception(e),
            ))                       
            raise 


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mito_ai.completions.prompt_builders.prompt_constants import (
    CHAT_CODE_FORMATTING_RULES,
    CITATION_RULES, 
    ACTIVE_CELL_ID_SECTION_HEADING, 
    CODE_SECTION_HEADING,
    get_database_rules
)

def create_chat_system_message_prompt() -> str:
    return f"""You are Mito Data Copilot, an AI assistant for Jupyter. You're a great python programmer, a seasoned data scientist and a subject matter expert.

The user is going to ask you for help writing code, debugging code, explaining code, or drawing conclusions from their data/graphs. It is your job to help them accomplish their goal. 

The user will give you a set of variables, existing code, and a task to complete. 

There are three possible types of responses you might give:
1. Code Update: If the task requires modifying or extending the existing code, respond with the updated active code cell and a short explanation of the changes made. 
2. Explanation/Analysis: If the task does not require a code update, it might instead require you to provide an explanation of existing code or data, provide an analysis of the the data or chart.
3. Friendly Response: If the user is just asking a question, saying hi, or you're just chatting, respond with a friendly response and do not return any code.

====
{CITATION_RULES}

<Example 1> 
{ACTIVE_CELL_ID_SECTION_HEADING}
'7b3a9e2c-5d14-4c83-b2f9-d67891e4a5f2'

{CODE_SECTION_HEADING}
```python
sales_df = pd.read_csv('sales_data.csv')
monthly_revenue = sales_df.groupby('month')['revenue'].sum()
top_month = monthly_revenue.idxmax()
peak_revenue = monthly_revenue.max()
growth_rate = (monthly_revenue.iloc[-1] / monthly_revenue.iloc[0] - 1) * 100
```

Your task: What are the key revenue insights from this sales data?

Output:
Peak monthly revenue reached $847,392 in March[MITO_CITATION:7b3a9e2c-5d14-4c83-b2f9-d67891e4a5f2:2-3], representing a 23.8% year-over-year growth rate[MITO_CITATION:7b3a9e2c-5d14-4c83-b2f9-d67891e4a5f2:4]. The revenue aggregation analysis[MITO_CITATION:7b3a9e2c-5d14-4c83-b2f9-d67891e4a5f2:1-2] reveals strong seasonal performance patterns.

</Example 1>

Notice in the example above:
- Citations support specific facts and numbers, not vague summaries
- Single line citations reference specific calculations (e.g., :4 for growth rate)
- Multiline citations reference broader analysis blocks (e.g., :1-2 for the groupby operation)
- Language is information-dense with concrete metrics
- All line numbers are 0-indexed

<Example 2>

{ACTIVE_CELL_ID_SECTION_HEADING}
'1a2b3c4d-5e6f-7g8h-9i0j-k1l2m3n4o5p6'

{CODE_SECTION_HEADING}
```python
```

Your task: Hello

Output:
Hey there! I'm Mito AI. How can I help you today? 

</Example 2>

Notice in the example above that the user is just sending a friendly message, so we respond with a friendly message and do not return any code.

===
{get_database_rules()}

==== 
{CHAT_CODE_FORMATTING_RULES}

IMPORTANT RULES:
- Do not recreate variables that already exist
- Keep as much of the original code as possible
- When updating an existing code cell, return the full code cell with the update applied. Do not only return part of the code cell with a comment like "# Updated code starts here", etc.
- Only update code in the active cell. Do not update other code in the notebook.
- Write code that preserves the intent of the original code shared with you and the task to complete.
- Make the solution as simple as possible.
- Reuse as much of the existing code as possible.
- Do not add temporary comments like '# Fixed the typo here' or '# Added this line to fix the error'
- Whenever writing Python code, it should be a python code block starting with ```python and ending with ```
"""



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
from datetime import datetime
from unittest.mock import patch
from mito_ai.utils.server_limits import (
    check_mito_server_quota,
    OS_MONTHLY_AI_COMPLETIONS_LIMIT,
    OS_MONTHLY_AUTOCOMPLETE_LIMIT,
)
from mito_ai.completions.models import MessageType
from mito_ai.utils.open_ai_utils import _prepare_request_data_and_headers

REALLY_OLD_DATE = "2020-01-01"
TODAY = datetime.now().strftime("%Y-%m-%d")


def test_check_mito_server_quota_open_source_user() -> None:
    # Under chat completions limit
    with patch("mito_ai.utils.server_limits.get_chat_completion_count", return_value=1) as mock_count, \
         patch("mito_ai.utils.server_limits.get_last_reset_date", return_value=TODAY) as mock_date, \
         patch("mito_ai.utils.server_limits.is_pro", return_value=False):
        
        check_mito_server_quota(MessageType.CHAT)
        assert mock_count.called
        assert mock_date.called
        assert mock_count.return_value == 1
        assert mock_date.return_value == TODAY

    # Under autocomplete limit
    with patch("mito_ai.utils.server_limits.get_autocomplete_count", return_value=1) as mock_count, \
         patch("mito_ai.utils.server_limits.get_last_reset_date", return_value=TODAY) as mock_date, \
         patch("mito_ai.utils.server_limits.is_pro", return_value=False):
        
        check_mito_server_quota(MessageType.INLINE_COMPLETION)
        assert mock_count.called
        assert mock_date.called
        assert mock_count.return_value == 1
        assert mock_date.return_value == TODAY

    # Over chat completions limit
    with pytest.raises(PermissionError), \
         patch("mito_ai.utils.server_limits.get_chat_completion_count", return_value=OS_MONTHLY_AI_COMPLETIONS_LIMIT + 1) as mock_count, \
         patch("mito_ai.utils.server_limits.get_last_reset_date", return_value=TODAY) as mock_date, \
         patch("mito_ai.utils.server_limits.is_pro", return_value=False):
        
        check_mito_server_quota(MessageType.CHAT)
        assert mock_count.called
        assert mock_date.called
        assert mock_count.return_value == OS_MONTHLY_AI_COMPLETIONS_LIMIT + 1
        assert mock_date.return_value == TODAY

    # Over autocomplete limit
    with pytest.raises(PermissionError), \
         patch("mito_ai.utils.server_limits.get_autocomplete_count", return_value=OS_MONTHLY_AUTOCOMPLETE_LIMIT + 1) as mock_count, \
         patch("mito_ai.utils.server_limits.get_last_reset_date", return_value=TODAY) as mock_date, \
         patch("mito_ai.utils.server_limits.is_pro", return_value=False):
        
        check_mito_server_quota(MessageType.INLINE_COMPLETION)
        assert mock_count.called
        assert mock_date.called
        assert mock_count.return_value == OS_MONTHLY_AUTOCOMPLETE_LIMIT + 1
        assert mock_date.return_value == TODAY


def test_check_mito_server_quota_pro_user() -> None:
    # No error should be thrown since pro users don't have limits
    with patch("mito_ai.utils.server_limits.is_pro", return_value=True), \
         patch("mito_ai.utils.server_limits.get_chat_completion_count", return_value=1000), \
         patch("mito_ai.utils.server_limits.get_last_reset_date", return_value=REALLY_OLD_DATE):
        
        check_mito_server_quota(MessageType.CHAT)

def test_prepare_request_data_and_headers_basic() -> None:
    """Test basic functionality of _prepare_request_data_and_headers"""

    # Mock the user fields
    with patch("mito_ai.utils.open_ai_utils.get_user_field") as mock_get_user_field:
        mock_get_user_field.side_effect = ["test@example.com", "user123"]
        
        # Mock the quota check
        with patch("mito_ai.utils.open_ai_utils.check_mito_server_quota") as mock_check_quota:
            data, headers = _prepare_request_data_and_headers(
                last_message_content="test message",
                ai_completion_data={"key": "value"},
                timeout=30,
                max_retries=3,
                message_type=MessageType.CHAT
            )

            # Verify quota check was called
            mock_check_quota.assert_called_once_with(MessageType.CHAT)

            # Verify data structure
            assert data["timeout"] == 30
            assert data["max_retries"] == 3
            assert data["email"] == "test@example.com"
            assert data["user_id"] == "user123"
            assert data["data"] == {"key": "value"}
            assert data["user_input"] == "test message"

            # Verify headers
            assert headers == {"Content-Type": "application/json"}

def test_prepare_request_data_and_headers_null_message() -> None:
    """Test handling of null message content"""
    with patch("mito_ai.utils.open_ai_utils.get_user_field") as mock_get_user_field:
        mock_get_user_field.side_effect = ["test@example.com", "user123"]
        
        with patch("mito_ai.utils.open_ai_utils.check_mito_server_quota"):
            data, _ = _prepare_request_data_and_headers(
                last_message_content=None,
                ai_completion_data={},
                timeout=30,
                max_retries=3,
                message_type=MessageType.CHAT
            )

            # Verify empty string is used for null message
            assert data["user_input"] == ""

def test_prepare_request_data_and_headers_caches_user_info() -> None:
    """Test that user info is cached after first call"""
    # Mock both the global variables and the get_user_field function
    with patch("mito_ai.utils.open_ai_utils.__user_email", None), \
         patch("mito_ai.utils.open_ai_utils.__user_id", None), \
         patch("mito_ai.utils.open_ai_utils.get_user_field") as mock_get_user_field:
        
        mock_get_user_field.side_effect = ["test@example.com", "user123"]
        
        with patch("mito_ai.utils.open_ai_utils.check_mito_server_quota"):
            # First call
            data1, _ = _prepare_request_data_and_headers(
                last_message_content="test",
                ai_completion_data={},
                timeout=30,
                max_retries=3,
                message_type=MessageType.CHAT
            )

            # Second call
            data2, _ = _prepare_request_data_and_headers(
                last_message_content="test",
                ai_completion_data={},
                timeout=30,
                max_retries=3,
                message_type=MessageType.CHAT
            )

            # Verify get_user_field was only called twice (once for email, once for user_id)
            assert mock_get_user_field.call_count == 2
            
            # Verify both calls return same user info
            assert data1["email"] == data2["email"] == "test@example.com"
            assert data1["user_id"] == data2["user_id"] == "user123"



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import logging
import time
import uuid
from dataclasses import asdict
from http import HTTPStatus
from typing import Any, Dict, Optional, Union
import tornado
import tornado.ioloop
import tornado.web
from jupyter_core.utils import ensure_async
from jupyter_server.base.handlers import JupyterHandler
from tornado.websocket import WebSocketHandler
from mito_ai.completions.message_history import GlobalMessageHistory
from mito_ai.logger import get_logger
from mito_ai.completions.models import (
    AgentSmartDebugMetadata,
    CompletionError,
    CompletionItem,
    CompletionReply,
    CompletionRequest,
    CompletionStreamChunk,
    ErrorMessage,
    FetchHistoryReply,
    StartNewChatReply,
    FetchThreadsReply,
    DeleteThreadReply,
    ChatMessageMetadata,
    SmartDebugMetadata,
    CodeExplainMetadata,
    AgentExecutionMetadata,
    InlineCompleterMetadata,
    MessageType
)
from mito_ai.completions.providers import OpenAIProvider
from mito_ai.utils.create import initialize_user
from mito_ai.utils.version_utils import is_pro
from mito_ai.completions.completion_handlers.chat_completion_handler import get_chat_completion, stream_chat_completion
from mito_ai.completions.completion_handlers.smart_debug_handler import get_smart_debug_completion, stream_smart_debug_completion
from mito_ai.completions.completion_handlers.code_explain_handler import get_code_explain_completion, stream_code_explain_completion
from mito_ai.completions.completion_handlers.inline_completer_handler import get_inline_completion
from mito_ai.completions.completion_handlers.agent_execution_handler import get_agent_execution_completion
from mito_ai.completions.completion_handlers.agent_auto_error_fixup_handler import get_agent_auto_error_fixup_completion
from mito_ai.utils.telemetry_utils import identify

FALLBACK_MODEL = "gpt-4.1"  # Default model to use for safety

# The GlobalMessageHistory is responsible for updating the message histories stored in the .mito/ai-chats directory.
# We create one GlobalMessageHistory per backend server instance instead of one per websocket connection so that the
# there is one manager of the locks for the .mito/ai-chats directory. This is my current understanding and it 
# might be incorrect!
message_history = GlobalMessageHistory()

# This handler is responsible for the mito_ai/completions endpoint.
# It takes a message from the user, sends it to the OpenAI API, and returns the response.
# Important: Because this is a server extension, print statements are sent to the
# jupyter server terminal by default (ie: the terminal you ran `jupyter lab`)
class CompletionHandler(JupyterHandler, WebSocketHandler):
    """Completion websocket handler."""

    def initialize(self, llm: OpenAIProvider) -> None:
        super().initialize()
        self.log.debug("Initializing websocket connection %s", self.request.path)
        self._llm = llm
        self.is_pro = is_pro()
        self._selected_model = FALLBACK_MODEL
        identify(llm.key_type)
        
    @property
    def log(self) -> logging.Logger:
        """Use Mito AI logger"""
        return get_logger()

    async def pre_get(self) -> None:
        """Handles websocket authentication/authorization."""
        # authenticate the request before opening the websocket
        user = self.current_user
        if user is None:
            self.log.warning("Couldn't authenticate WebSocket connection")
            raise tornado.web.HTTPError(HTTPStatus.UNAUTHORIZED)

        # authorize the user.
        if not await ensure_async(
            self.authorizer.is_authorized(self, user, "execute", "mito_ai-completion")
        ):
            raise tornado.web.HTTPError(HTTPStatus.FORBIDDEN)

    async def get(self, *args: Any, **kwargs: Dict[str, Any]) -> None:
        """Get an event to open a socket or check service availability."""
        # Check if this is just a service availability check
        if self.get_query_argument('check_availability', None) == 'true':
            self.set_status(HTTPStatus.OK)
            self.finish()
            return

        await ensure_async(self.pre_get()) # type: ignore

        initialize_user()

        reply = super().get(*args, **kwargs)
        if reply is not None:
            await reply

    def on_close(self) -> None:
        """Invoked when the WebSocket is closed.

        If the connection was closed cleanly and a status code or reason
        phrase was supplied, these values will be available as the attributes
        ``self.close_code`` and ``self.close_reason``.
        """
        # Stop observing the provider error
        self._llm.unobserve(self._send_error, "last_error")
    
        
    async def on_message(self, message: str) -> None: # type: ignore
        """Handle incoming messages on the WebSocket.

        Args:
            message: The message received on the WebSocket.
        """
        start = time.time()
        self.log.debug("Message received: %s", message)
        
        try:
            parsed_message = json.loads(message)
            metadata_dict = parsed_message.get('metadata', {})
            type: MessageType = MessageType(parsed_message.get('type'))
        except ValueError as e:
            self.log.error("Invalid completion request.", exc_info=e)
            return
        
        reply: Union[StartNewChatReply, FetchThreadsReply, DeleteThreadReply, FetchHistoryReply, CompletionReply]

        # Clear history if the type is "start_new_chat"
        if type == MessageType.START_NEW_CHAT:
            thread_id = message_history.create_new_thread()
            
            reply = StartNewChatReply(
                parent_id=parsed_message.get("message_id"),
                thread_id=thread_id
            )
            self.reply(reply)
            return

        # Handle get_threads: return list of chat threads
        if type == MessageType.GET_THREADS:
            threads = message_history.get_threads()
            reply = FetchThreadsReply(
                parent_id=parsed_message.get("message_id"),
                threads=threads
            )
            self.reply(reply)
            return

        # Handle delete_thread: delete the specified thread
        if type == MessageType.DELETE_THREAD:
            thread_id_to_delete = metadata_dict.get('thread_id')
            if thread_id_to_delete:
                is_thread_deleted = message_history.delete_thread(thread_id_to_delete)
                reply = DeleteThreadReply(
                    parent_id=parsed_message.get("message_id"),
                    success=is_thread_deleted
                )
            else:
                reply = DeleteThreadReply(
                    parent_id=parsed_message.get("message_id"),
                    success=False
                )
            self.reply(reply)
            return
        if type == MessageType.FETCH_HISTORY:
            
            # If a thread_id is provided, use that thread's history; otherwise, use newest.
            thread_id = metadata_dict.get('thread_id')
            display_history = message_history.get_display_history(thread_id)
            
            reply = FetchHistoryReply(
                parent_id=parsed_message.get('message_id'),
                items=display_history
            )
            self.reply(reply)
            return
        
        # Updated handler for receiving model selection via websocket
        if type == MessageType.UPDATE_MODEL_CONFIG:
            model = metadata_dict.get('model')
            if model:
                self._selected_model = model
                self.log.info(f"Model updated to: {model}")
                reply = CompletionReply(
                    items=[CompletionItem(content=f"Model updated to {model}", isIncomplete=False)],
                    parent_id=parsed_message.get('message_id')
                )
                self.reply(reply)
            else:
                error = CompletionError(
                    error_type="InvalidModelConfig",
                    title="Invalid model configuration",
                    traceback="",
                    hint="Model name is required"
                )
                reply = CompletionReply(
                    items=[],
                    error=error,
                    parent_id=parsed_message.get('message_id')
                )
                self.reply(reply)
            return
        
        try:
            # Get completion based on message type
            completion = None
            message_id = parsed_message.get('message_id')
            stream = parsed_message.get('stream')

            # When handling completions, always use the selected model
            model = self._selected_model
            if type == MessageType.CHAT:
                chat_metadata = ChatMessageMetadata(**metadata_dict)
                
                # Handle streaming if requested and available
                if stream:
                    # Use stream_chat_completion to stream the response
                    await stream_chat_completion(
                        chat_metadata, 
                        self._llm, 
                        message_history, 
                        message_id,
                        self.reply,
                        model
                    )
                    return
                else:
                    # Regular non-streaming completion
                    completion = await get_chat_completion(chat_metadata, self._llm, message_history, model)
            elif type == MessageType.SMART_DEBUG:
                smart_debug_metadata = SmartDebugMetadata(**metadata_dict)
                # Handle streaming if requested and available
                if stream:
                    # Use stream_smart_debug_completion to stream the response
                    await stream_smart_debug_completion(
                        smart_debug_metadata, 
                        self._llm, 
                        message_history, 
                        message_id,
                        self.reply,
                        model
                    )
                    return
                else:
                    # Regular non-streaming completion
                    completion = await get_smart_debug_completion(smart_debug_metadata, self._llm, message_history, model)
            elif type == MessageType.CODE_EXPLAIN:
                code_explain_metadata = CodeExplainMetadata(**metadata_dict)

                # Handle streaming if requested and available
                if stream:
                    # Use stream_code_explain_completion to stream the response
                    await stream_code_explain_completion(
                        code_explain_metadata, 
                        self._llm, 
                        message_history,
                        message_id,
                        self.reply,
                        model
                    )
                    return
                else:
                    # Regular non-streaming completion
                    completion = await get_code_explain_completion(code_explain_metadata, self._llm, message_history, model)
            elif type == MessageType.AGENT_EXECUTION:
                agent_execution_metadata = AgentExecutionMetadata(**metadata_dict)
                completion = await get_agent_execution_completion(agent_execution_metadata, self._llm, message_history, model)
            elif type == MessageType.AGENT_AUTO_ERROR_FIXUP:
                agent_auto_error_fixup_metadata = AgentSmartDebugMetadata(**metadata_dict)
                completion = await get_agent_auto_error_fixup_completion(agent_auto_error_fixup_metadata, self._llm, message_history, model)
            elif type == MessageType.INLINE_COMPLETION:
                inline_completer_metadata = InlineCompleterMetadata(**metadata_dict)
                completion = await get_inline_completion(inline_completer_metadata, self._llm, message_history, model)
            else:
                raise ValueError(f"Invalid message type: {type}")
            
            # Create and send reply
            reply = CompletionReply(
                items=[CompletionItem(content=completion, isIncomplete=False)],
                parent_id=message_id
            )
            self.reply(reply)
            
            latency_ms = round((time.time() - start) * 1000)
            self.log.info(f"Completion handler resolved in {latency_ms} ms.")

        except Exception as e:
            error = CompletionError.from_exception(e)
            self._send_error({"new": error})
            reply = CompletionReply(
                items=[],
                error=error,
                parent_id=parsed_message.get('message_id')
            )
            self.reply(reply)

    def open(self, *args: str, **kwargs: str) -> None:
        """Invoked when a new WebSocket is opened.

        The arguments to `open` are extracted from the `tornado.web.URLSpec`
        regular expression, just like the arguments to
        `tornado.web.RequestHandler.get`.

        `open` may be a coroutine. `on_message` will not be called until
        `open` has returned.
        """
        if self._llm.last_error:
            self._send_error({"new": self._llm.last_error})
        # Start observing the provider error
        self._llm.observe(self._send_error, "last_error")
        # Send the server capabilities to the client.
        self.reply(self._llm.capabilities)
        

    async def handle_exception(self, e: Exception, request: CompletionRequest) -> None:
        """
        Handles an exception raised in either ``handle_request`` or
        ``handle_stream_request``.

        Args:
            e: The exception raised.
            request: The completion request that caused the exception.
        """
        hint = ""
        if isinstance(e, PermissionError):
            hint = "You've reached the free tier limit for Mito AI. Upgrade to Pro for unlimited uses or supply your own OpenAI API key."
        elif "openai" in self._llm.capabilities.provider.lower():
            hint = "There was an error communicating with OpenAI. This might be due to a temporary OpenAI outage, a problem with your internet connection, or an incorrect API key. Please try again."
        else:
            hint = "There was an error communicating with Mito server. This might be due to a temporary server outage or a problem with your internet connection. Please try again."
        
        error: CompletionError = CompletionError.from_exception(e, hint=hint)
        self._send_error({"new": error})
        
        reply: Union[CompletionStreamChunk, CompletionReply]
        if request.stream:
            reply = CompletionStreamChunk(
                chunk=CompletionItem(content="", isIncomplete=True),
                parent_id=request.message_id,
                done=True,
                error=error,
            )
        else:
            reply = CompletionReply(
                items=[],
                error=error,
                parent_id=request.message_id,
            )
        self.reply(reply)
        
    def reply(self, reply: Any) -> None:
        """Write a reply object to the WebSocket connection.

        Args:
            reply: The completion reply object.
                It must be a dataclass instance.
        """
        message = asdict(reply)
        super().write_message(message)

    def _send_error(self, change: Dict[str, Any]) -> None:
        """Send an error message to the client."""
        error = change["new"]

        self.reply(
            ErrorMessage(**asdict(error))
            if error is not None
            else ErrorMessage(error_type="", title="No error", traceback="")
        )



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import logging
from dataclasses import asdict
from http import HTTPStatus
from typing import Any, Dict, Optional

import tornado
from jupyter_core.utils import ensure_async
from jupyter_server.base.handlers import JupyterHandler
from tornado.websocket import WebSocketHandler

from mito_ai.logger import get_logger


class BaseWebSocketHandler(JupyterHandler, WebSocketHandler):
    """Base WebSocket handler with common functionality for Mito AI services."""

    def initialize(self) -> None:
        """Initialize the WebSocket handler."""
        super().initialize()
        self.log.debug("Initializing websocket connection %s", self.request.path)

    @property
    def log(self) -> logging.Logger:
        """Use Mito AI logger."""
        return get_logger()

    async def pre_get(self) -> None:
        """Handle WebSocket authentication/authorization."""
        # Authenticate the request before opening the WebSocket
        user = self.current_user
        if user is None:
            self.log.warning("Couldn't authenticate WebSocket connection")
            raise tornado.web.HTTPError(HTTPStatus.UNAUTHORIZED)

        # Authorize the user
        if not await ensure_async(
            self.authorizer.is_authorized(self, user, "execute", "mito-ai")
        ):
            raise tornado.web.HTTPError(HTTPStatus.FORBIDDEN)

    async def get(self, *args: Any, **kwargs: Dict[str, Any]) -> None:
        """Get an event to open a socket or check service availability."""
        # Check if this is just a service availability check
        if self.get_query_argument('check_availability', None) == 'true':
            self.set_status(HTTPStatus.OK)
            self.finish()
            return

        await ensure_async(self.pre_get())  # type: ignore

        reply = super().get(*args, **kwargs)
        if reply is not None:
            await reply

    def reply(self, reply: Any) -> None:
        """Write a reply object to the WebSocket connection.

        Args:
            reply: The reply object, which must be a dataclass instance.
        """
        message = asdict(reply)
        super().write_message(message)

    def on_close(self) -> None:
        """Invoked when the WebSocket is closed."""
        self.log.debug("WebSocket closed: %s", self.request.path)

    def parse_message(self, message: str) -> Dict[str, Any]:
        """Parse an incoming message.
        
        Args:
            message: The JSON message string to parse.
            
        Returns:
            The parsed message as a dict.
            
        Raises:
            ValueError: If the message is not valid JSON.
        """
        try:
            return json.loads(message) # type: ignore
        except ValueError as e:
            self.log.error("Invalid message: %s", e)
            raise 